@@ -7,8 +7,7 @@ License     : ISC
 Maintainer  : emertens@gmail.com
 
 Implementation of ECDSA-NIST256P-CHALLENGE SASL authentication mode
-as implemented at <https://github.com/kaniini/ecdsatool> and used
-on Freenode.
+as implemented at <https://github.com/kaniini/ecdsatool>.
 
 Using this mode requires that the @ecdsa@ utility program is available
 in your search path.
@@ -38,11 +37,13 @@ authenticationMode = "ECDSA-NIST256P-CHALLENGE"
 
 -- | Encode a username as specified in this authentication mode.
 encodeAuthentication ::
-  Text {- ^ authorization identity  -} ->
+  Maybe Text {- ^ authorization identity  -} ->
   Text {- ^ authentication identity -} ->
   AuthenticatePayload
-encodeAuthentication authz authc =
-  AuthenticatePayload (Text.encodeUtf8 (authz <> "\0" <> authc <> "\0"))
+encodeAuthentication Nothing authc =
+  AuthenticatePayload (Text.encodeUtf8 authc)
+encodeAuthentication (Just authz) authc =
+  AuthenticatePayload (Text.encodeUtf8 (authc <> "\0" <> authz))
 
 
 -- | Compute the response for a given challenge using the @ecdsatool@
