@@ -1,4 +1,4 @@
-{-# Language RecordWildCards #-}
+{-# Language OverloadedStrings, RecordWildCards #-}
 {-|
 Module      : Client.CApi.Exports
 Description : Foreign exports which expose functionality for extensions
@@ -82,9 +82,14 @@ module Client.CApi.Exports
  , Glirc_cancel_timer
  , glirc_cancel_timer
 
+ , Glirc_window_lines
+ , glirc_window_lines
+
+ , Glirc_thread
+ , glirc_thread
  ) where
 
-import           Client.CApi (cancelTimer, pushTimer)
+import           Client.CApi (cancelTimer, pushTimer, ThreadEntry(..), ActiveExtension(aeThreads))
 import           Client.CApi.Types
 import           Client.Configuration
 import           Client.Message
@@ -92,9 +97,11 @@ import           Client.State
 import           Client.State.Channel
 import           Client.State.Focus
 import           Client.State.Network
-import           Client.State.Window
+import           Client.State.Window (windowClear, windowSeen, winMessages, wlText)
 import           Client.UserHost
+import           Control.Concurrent (forkOS)
 import           Control.Concurrent.MVar
+import           Control.Concurrent.STM (atomically, writeTQueue)
 import           Control.Exception
 import           Control.Lens
 import           Control.Monad (unless)
@@ -106,6 +113,7 @@ import           Data.Monoid (First(..))
 import qualified Data.HashMap.Strict as HashMap
 import           Data.Text (Text)
 import qualified Data.Text as Text
+import qualified Data.Text.Lazy as LText
 import qualified Data.Text.Foreign as Text
 import           Data.Time
 import           Foreign.C
@@ -240,7 +248,7 @@ glirc_inject_chat stab netPtr netLen srcPtr srcLen tgtPtr tgtLen msgPtr msgLen =
      now  <- getZonedTime
 
      let msg = ClientMessage
-                 { _msgBody    = IrcBody (Privmsg (parseUserInfo src) tgt txt)
+                 { _msgBody    = IrcBody (Privmsg (Source (parseUserInfo src) "") tgt txt)
                  , _msgTime    = now
                  , _msgNetwork = net
                  }
@@ -555,7 +563,7 @@ glirc_clear_window stab networkPtr networkLen channelPtr channelLen =
 
      mvar <- derefToken stab
      modifyMVar_ mvar $ \(i,st) ->
-       let st' = set (clientWindows . ix focus) emptyWindow st
+       let st' = over (clientWindows . ix focus) windowClear st
        in st' `seq` return (i,st')
 
 ------------------------------------------------------------------------
@@ -764,3 +772,62 @@ glirc_cancel_timer stab timerId =
        in return $! case mb of
             Just (First (Just ptr), st') -> ((i,st'), ptr)
             _ -> ((i, st), nullPtr)
+
+------------------------------------------------------------------------
+
+-- | Type of 'glirc_window_lines' extension entry-point
+type Glirc_window_lines =
+  Ptr ()  {- ^ api token       -} ->
+  CString {- ^ network name    -} ->
+  CSize   {- ^ network length  -} ->
+  CString {- ^ target name     -} ->
+  CSize   {- ^ target length   -} ->
+  CInt    {- ^ use filter      -} ->
+  IO (Ptr CString) {- ^ null terminated array of null terminated strings -}
+
+-- | This extension entry-point allocates a list of all the window lines for
+-- the requested window. The lines are presented with newest line at the head
+-- of the list.
+-- The caller is responsible for freeing successful result with
+-- @glirc_free_strings@.
+glirc_window_lines :: Glirc_window_lines
+glirc_window_lines stab net netL tgt tgtL filt =
+  do mvar <- derefToken stab
+     (_,st) <- readMVar mvar
+     network <- peekFgnStringLen (FgnStringLen net netL)
+     channel <- peekFgnStringLen (FgnStringLen tgt tgtL)
+     let focus
+           | Text.null network = Unfocused
+           | Text.null channel = NetworkFocus network
+           | otherwise         = ChannelFocus network (mkId channel)
+         filterFun
+           | filt == 0 = id
+           | otherwise = clientFilter st id
+         strs = toListOf (clientWindows . ix focus . winMessages . each . wlText) st
+     ptrs <- traverse (newCString . LText.unpack) (filterFun strs)
+     newArray0 nullPtr ptrs
+
+------------------------------------------------------------------------
+
+-- | Type of 'glirc_thread' extension entry-point
+type Glirc_thread =
+  Ptr ()  {- ^ api token -} ->
+  FunPtr (Ptr () -> IO (Ptr ())) {- ^ start -} ->
+  FunPtr (Ptr () -> IO ()) {- ^ finish -} ->
+  Ptr () {- ^ start argument  -} ->
+  IO ()  {- ^ null terminated array of null terminated strings -}
+
+-- | This extension entry-point allocates a list of all the window lines for
+-- the requested window. The lines are presented with newest line at the head
+-- of the list.
+-- The caller is responsible for freeing successful result with
+-- @glirc_free_strings@.
+glirc_thread :: Glirc_thread
+glirc_thread stab start finish arg =
+  do mvar <- derefToken stab
+     modifyMVar_ mvar $ \(i,st) ->
+       do _ <- forkOS $
+            do result <- runThreadStart start arg
+               atomically (writeTQueue (view clientThreadJoins st) (i, ThreadEntry finish result))
+          let incThreads ae = ae { aeThreads = aeThreads ae + 1}
+          pure (i, over (clientExtensions . esActive . ix i) incThreads st)
