@@ -32,7 +32,6 @@ import           Client.Commands.Interpolation
 import           Client.Commands.Recognizer
 import           Client.Commands.WordCompletion
 import           Client.Configuration
-import           Client.Message
 import           Client.State
 import           Client.State.Extensions
 import           Client.State.Focus
@@ -45,7 +44,7 @@ import           Control.Monad
 import           Data.Foldable
 import           Data.Text (Text)
 import qualified Data.Text as Text
-import           Data.Time (ZonedTime, getZonedTime)
+import           Data.Time (getZonedTime)
 import           Irc.Commands
 import           Irc.Identifier
 import           Irc.RawIrcMsg
@@ -54,9 +53,9 @@ import           RtsStats (getStats)
 import           System.Process
 
 import           Client.Commands.Channel (channelCommands)
+import           Client.Commands.Certificate (newCertificateCommand)
 import           Client.Commands.Chat (chatCommands, chatCommand', executeChat)
 import           Client.Commands.Connection (connectionCommands)
-import           Client.Commands.DCC (dccCommands)
 import           Client.Commands.Operator (operatorCommands)
 import           Client.Commands.Queries (queryCommands)
 import           Client.Commands.Toggles (togglesCommands)
@@ -95,7 +94,7 @@ executeUserCommand discoTime command st = do
       rest = dropWhile (==' ') (dropWhile (/=' ') command)
 
   case views (clientConfig . configMacros) (recognize key) st of
-    Exact (Macro (MacroSpec spec) cmdExs) ->
+    Exact (Macro _ (MacroSpec spec) cmdExs) ->
       case doExpansion spec cmdExs rest of
         Nothing   -> commandFailureMsg "macro expansions failed" st
         Just cmds -> process cmds st
@@ -317,6 +316,8 @@ commandsList =
       \ URL to open counting back from the most recent.\n"
     $ ClientCommand cmdUrl noClientTab
 
+  , newCertificateCommand
+
   , Command
       (pure "help")
       (optionalArg (simpleToken "[command]"))
@@ -330,8 +331,8 @@ commandsList =
   ],
 
   togglesCommands, connectionCommands, windowCommands, chatCommands,
-  queryCommands, channelCommands, zncCommands, operatorCommands,
-  dccCommands ]
+  queryCommands, channelCommands, zncCommands, operatorCommands
+  ]
 
 -- | Implementation of @/exit@ command.
 cmdExit :: ClientCommand ()
@@ -404,12 +405,13 @@ tabReload _ st _ = commandFailure st
 commandNameCompletion :: Bool -> ClientState -> Maybe ClientState
 commandNameCompletion isReversed st =
   do guard (cursorPos == n)
-     clientTextBox (wordComplete plainWordCompleteMode isReversed [] possibilities) st
+     clientTextBox (wordComplete (' ' /=) plainWordCompleteMode isReversed [] possibilities) st
   where
     n = length white + length leadingPart
     (cursorPos, line) = clientLine st
     (white, leadingPart) = takeWhile (' ' /=) <$> span (' '==) line
-    possibilities = Text.cons '/' <$> commandNames
+
+    possibilities = caseText . Text.cons '/' <$> commandNames
     commandNames = keys commands
                 ++ keys (view (clientConfig . configMacros) st)
 
@@ -488,7 +490,7 @@ cmdExec st rest =
         do sendMsg cs (ircPrivmsg channel msg)
            chatCommand'
               (\src tgt -> Privmsg src tgt msg)
-              channel
+              [channel]
               cs st1) st (filter (not . Text.null) msgs)
 
     currentNetworkState =
@@ -498,14 +500,6 @@ cmdExec st rest =
     failure now es =
       commandFailure $! foldl' (flip (recordError now "")) st (map Text.pack es)
 
-recordSuccess :: ZonedTime -> ClientState -> Text -> ClientState
-recordSuccess now ste m =
-  recordNetworkMessage ClientMessage
-    { _msgTime    = now
-    , _msgBody    = NormalBody m
-    , _msgNetwork = ""
-    } ste
-
 
 cmdUrl :: ClientCommand (Maybe Int)
 cmdUrl st arg =
@@ -523,9 +517,11 @@ cmdUrl st arg =
         Just url -> openUrl opener (Text.unpack url) st
         Nothing  -> commandFailureMsg "bad url number" st
 
-openUrl :: FilePath -> String -> ClientState -> IO CommandResult
-openUrl opener url st =
-  do res <- try (callProcess opener [url])
+openUrl :: UrlOpener -> String -> ClientState -> IO CommandResult
+openUrl (UrlOpener opener args) url st =
+  do let argStr (UrlArgLiteral str) = str
+         argStr UrlArgUrl           = url
+     res <- try (callProcess opener (map argStr args))
      case res of
        Left e  -> commandFailureMsg (Text.pack (displayException (e :: IOError))) st
        Right{} -> commandSuccess st
