@@ -31,7 +31,6 @@ import           Irc.Commands
 import           Irc.Identifier
 import           Irc.Message
 import           Irc.RawIrcMsg
-import           Irc.UserInfo
 
 chatCommands :: CommandSection
 chatCommands = CommandSection "IRC commands"
@@ -198,6 +197,44 @@ chatCommands = CommandSection "IRC commands"
       \\^BSee also:\^B me, msg, say\n"
     $ NetworkCommand cmdNotice simpleNetworkTab
 
+  , Command
+      (pure "wallops")
+      (remainingArg "message to +w users")
+      "\^BParameters:\^B\n\
+      \\n\
+      \    message: Formatted message body\n\
+      \\n\
+      \\^BDescription:\^B\n\
+      \\n\
+      \    Send a network-wide WALLOPS message. These message go out\n\
+      \    to users who have the 'w' usermode set.\n\
+      \\n\
+      \\^BExamples:\^B\n\
+      \\n\
+      \    /wallops Hi everyone, thanks for using this network!\n\
+      \\n\
+      \\^BSee also:\^B me, msg, say\n"
+    $ NetworkCommand cmdWallops simpleNetworkTab
+
+  , Command
+      (pure "operwall")
+      (remainingArg "message to +z opers")
+      "\^BParameters:\^B\n\
+      \\n\
+      \    message: Formatted message body\n\
+      \\n\
+      \\^BDescription:\^B\n\
+      \\n\
+      \    Send a network-wide WALLOPS message to opers. These message go\n\
+      \    out to opers who have the 'z' usermode set.\n\
+      \\n\
+      \\^BExamples:\^B\n\
+      \\n\
+      \    /operwall What's this even for?\n\
+      \\n\
+      \\^BSee also:\^B me, msg, say\n"
+    $ NetworkCommand cmdOperwall simpleNetworkTab
+
   , Command
       (pure "ctcp")
       (liftA3 (,,) (simpleToken "target") (simpleToken "command") (remainingArg "arguments"))
@@ -260,7 +297,7 @@ chatCommands = CommandSection "IRC commands"
     $ NetworkCommand cmdAway simpleNetworkTab
 
   , Command
-      ("users" :| ["names"])
+      (pure "names")
       (pure ())
       "\^BDescription:\^B\n\
       \\n\
@@ -270,7 +307,7 @@ chatCommands = CommandSection "IRC commands"
       \    Press ESC to exit the userlist.\n\
       \\n\
       \\^BSee also:\^B channelinfo, masks\n"
-    $ ChannelCommand cmdUsers  noChannelTab
+    $ ChannelCommand cmdChanNames noChannelTab
 
   , Command
       (pure "channelinfo")
@@ -333,8 +370,8 @@ cmdMonitor cs st args =
   do sendMsg cs (ircMonitor (fmap Text.pack args))
      commandSuccess st
 
-cmdUsers :: ChannelCommand ()
-cmdUsers _ _ st _ = commandSuccess (changeSubfocus FocusUsers st)
+cmdChanNames :: ChannelCommand ()
+cmdChanNames _ _ st _ = commandSuccess (changeSubfocus FocusUsers st)
 
 cmdChannelInfo :: ChannelCommand ()
 cmdChannelInfo _ _ st _ = commandSuccess (changeSubfocus FocusInfo st)
@@ -364,10 +401,10 @@ cmdQuery st (target, msg) =
       | Just cs <- preview (clientConnection net) st ->
            do let tgtTxt = idText tgt
                   msgTxt = Text.pack msg
-              sendMsg cs (ircPrivmsg tgtTxt msgTxt)
               chatCommand
-                 (\src tgt1 -> Privmsg src tgt1 msgTxt)
-                 tgtTxt cs st'
+                (ircPrivmsg tgtTxt msgTxt)
+                (\src tgt1 -> Privmsg src tgt1 msgTxt)
+                tgtTxt cs st'
       where
        firstTgt = mkId (Text.takeWhile (','/=) (idText tgt))
        st' = changeFocus (ChannelFocus net firstTgt) st
@@ -377,63 +414,90 @@ cmdQuery st (target, msg) =
 -- | Implementation of @/ctcp@
 cmdCtcp :: NetworkCommand (String, String, String)
 cmdCtcp cs st (target, cmd, args) =
-  do let cmdTxt = Text.toUpper (Text.pack cmd)
-         argTxt = Text.pack args
-         tgtTxt = Text.pack target
+ do let cmdTxt = Text.toUpper (Text.pack cmd)
+        argTxt = Text.pack args
+        tgtTxt = Text.pack target
 
-     sendMsg cs (ircPrivmsg tgtTxt ("\^A" <> cmdTxt <> " " <> argTxt <> "\^A"))
-     chatCommand
-        (\src tgt -> Ctcp src tgt cmdTxt argTxt)
-        tgtTxt cs st
+    let msg = "\^A" <> cmdTxt <>
+              (if Text.null argTxt then "" else " " <> argTxt) <>
+              "\^A"
+    chatCommand
+      (ircPrivmsg tgtTxt msg)
+      (\src tgt -> Ctcp src tgt cmdTxt argTxt)
+      tgtTxt cs st
+
+-- | Implementation of @/wallops@
+cmdWallops :: NetworkCommand String
+cmdWallops cs st rest
+  | null rest = commandFailureMsg "empty message" st
+  | otherwise =
+      do let restTxt = Text.pack rest
+         sendMsg cs (ircWallops restTxt)
+         commandSuccess st
+
+-- | Implementation of @/operwall@
+cmdOperwall :: NetworkCommand String
+cmdOperwall cs st rest
+  | null rest = commandFailureMsg "empty message" st
+  | otherwise =
+      do let restTxt = Text.pack rest
+         sendMsg cs (ircOperwall restTxt)
+         commandSuccess st
 
 -- | Implementation of @/notice@
 cmdNotice :: NetworkCommand (String, String)
 cmdNotice cs st (target, rest)
   | null rest = commandFailureMsg "empty message" st
   | otherwise =
-      do let restTxt = Text.pack rest
-             tgtTxt = Text.pack target
-
-         sendMsg cs (ircNotice tgtTxt restTxt)
-         chatCommand
-            (\src tgt -> Notice src tgt restTxt)
-            tgtTxt cs st
+     do let restTxt = Text.pack rest
+            tgtTxt = Text.pack target
+        chatCommand
+          (ircNotice tgtTxt restTxt)
+          (\src tgt -> Notice src tgt restTxt)
+          tgtTxt cs st
 
 -- | Implementation of @/msg@
 cmdMsg :: NetworkCommand (String, String)
 cmdMsg cs st (target, rest)
   | null rest = commandFailureMsg "empty message" st
   | otherwise =
-      do let restTxt = Text.pack rest
-             tgtTxt = Text.pack target
+     do let restTxt = Text.pack rest
+            tgtTxt = Text.pack target
+        chatCommand
+          (ircPrivmsg tgtTxt restTxt)
+          (\src tgt -> Privmsg src tgt restTxt)
+          tgtTxt cs st
+        
 
-         sendMsg cs (ircPrivmsg tgtTxt restTxt)
-         chatCommand
-            (\src tgt -> Privmsg src tgt restTxt)
-            tgtTxt cs st
 
 -- | Common logic for @/msg@ and @/notice@
 chatCommand ::
-  (UserInfo -> Identifier -> IrcMsg) ->
-  Text {- ^ target  -} ->
+  RawIrcMsg {- ^ irc command -} ->
+  (Source -> Identifier -> IrcMsg) ->
+  Text {- ^ targets -} ->
   NetworkState         ->
   ClientState          ->
   IO CommandResult
-chatCommand mkmsg target cs st =
-  commandSuccess =<< chatCommand' mkmsg target cs st
+chatCommand ircMsg mkmsg tgtsTxt cs st
+  | any Text.null tgtTxts = commandFailureMsg "empty target" st
+  | otherwise =
+   do sendMsg cs ircMsg
+      st' <- chatCommand' mkmsg tgtTxts cs st
+      commandSuccess st'
+  where
+    tgtTxts = Text.split (','==) tgtsTxt
 
 -- | Common logic for @/msg@ and @/notice@ returning the client state
 chatCommand' ::
-  (UserInfo -> Identifier -> IrcMsg) ->
-  Text {- ^ target  -} ->
+  (Source -> Identifier -> IrcMsg) ->
+  [Text] {- ^ targets  -} ->
   NetworkState         ->
   ClientState          ->
   IO ClientState
-chatCommand' con targetsTxt cs st =
+chatCommand' con targetTxts cs st =
   do now <- getZonedTime
-     let targetTxts = Text.split (==',') targetsTxt
-         targetIds  = mkId <$> targetTxts
-         !myNick = UserInfo (view csNick cs) "" ""
+     let targetIds = mkId <$> targetTxts
+         !myNick = Source (view csUserInfo cs) ""
          network = view csNetwork cs
          entries = [ (targetId,
                           ClientMessage
@@ -484,7 +548,7 @@ cmdMe :: ChannelCommand String
 cmdMe channelId cs st rest =
   do now <- getZonedTime
      let actionTxt = Text.pack ("\^AACTION " ++ rest ++ "\^A")
-         !myNick = UserInfo (view csNick cs) "" ""
+         !myNick = Source (view csUserInfo cs) ""
          network = view csNetwork cs
          entry = ClientMessage
                     { _msgTime = now
@@ -512,7 +576,7 @@ executeChat msg st =
 
              when allow (sendMsg cs (ircPrivmsg tgtTxt msgTxt))
 
-             let myNick = UserInfo (view csNick cs) "" ""
+             let myNick = Source (view csUserInfo cs) ""
                  entry = ClientMessage
                    { _msgTime    = now
                    , _msgNetwork = network
