@@ -13,12 +13,11 @@ import           Client.Commands.Arguments.Spec
 import           Client.Commands.TabCompletion
 import           Client.Commands.Types
 import           Client.Commands.WordCompletion
-import           Client.Image.PackedImage
 import           Client.Mask (buildMask)
 import           Client.State
 import           Client.State.Focus
 import           Client.State.Network
-import           Client.State.Window (emptyWindow, WindowLines((:-), Nil), wlFullImage, winMessages)
+import           Client.State.Window (windowClear, wlText, winMessages, winHidden, winSilent, winName)
 import           Control.Applicative
 import           Control.Exception
 import           Control.Lens
@@ -198,11 +197,13 @@ windowCommands = CommandSection "Window management"
       "Set the persistent regular expression.\n\
       \\n\
       \\^BFlags:\^B\n\
-      \    -An  Show n messages after match\n\
-      \    -Bn  Show n messages before match\n\
-      \    -Cn  Show n messages before and after match\n\
+      \    -A n Show n messages after match\n\
+      \    -B n Show n messages before match\n\
+      \    -C n Show n messages before and after match\n\
+      \    -F   Use plain-text match instead of regular expression\n\
       \    -i   Case insensitive match\n\
       \    -v   Invert pattern match\n\
+      \    -m n Limit results to n matches\n\
       \    --   Stop processing flags\n\
       \\n\
       \Clear the regular expression by calling this without an argument.\n\
@@ -225,8 +226,76 @@ windowCommands = CommandSection "Window management"
       \the regular expression instead.\n"
     $ ClientCommand cmdMentions noClientTab
 
+  , Command
+      (pure "setwindow")
+      (simpleToken "hide|show|loud|silent")
+      "Set window property.\n\
+      \\n\
+      \\^Bloud\^B / \^Bsilent\^B\n\
+      \    Toggles if window activity appears in the status bar.\n\
+      \n\
+      \\^Bshow\^B / \^Bhide\^B\n\
+      \    Toggles if window appears in window command shortcuts.\n"
+    $ ClientCommand cmdSetWindow tabSetWindow
+
+  , Command
+      (pure "setname")
+      (optionalArg (simpleToken "[letter]"))
+      "Set window shortcut letter. If no letter is provided the next available\n\
+      \letter will automatically be assigned.\n\
+      \\n\
+      \Available letters are configured in the 'window-names' configuration setting.\n"
+    $ ClientCommand cmdSetWindowName noClientTab
+
   ]
 
+cmdSetWindowName :: ClientCommand (Maybe String)
+cmdSetWindowName st arg =
+  -- unset current name so that it becomes available
+  let mbSt1 = failover (clientWindows . ix (view clientFocus st) . winName) (\_ -> Nothing) st in
+  case mbSt1 of
+    Nothing -> commandFailureMsg "no current window" st
+    Just st1 ->
+      let next = clientNextWindowName st
+          mbName =
+            case arg of
+              Just [n] | n `elem` clientWindowNames st -> Right n
+              Just _ -> Left "invalid name"
+              Nothing
+                | next /= '\0' -> Right next
+                | otherwise -> Left "no free names" in
+      case mbName of
+        Left e -> commandFailureMsg e st
+        Right name ->
+          let unset n = if n == Just name then Nothing else n in
+          commandSuccess
+            $ set  (clientWindows . ix (view clientFocus st) . winName) (Just name)
+            $ over (clientWindows . each                     . winName) unset
+            $ st1
+
+cmdSetWindow :: ClientCommand String
+cmdSetWindow st cmd =
+  case mbFun of
+    Nothing -> commandFailureMsg "bad window setting" st
+    Just f ->
+      case failover (clientWindows . ix (view clientFocus st)) f st of
+        Nothing -> commandFailureMsg "no such window" st
+        Just st' -> commandSuccess st'
+  where
+    mbFun =
+      case cmd of
+        "show"   -> Just (set winHidden False)
+        "hide"   -> Just (set winName Nothing . set winHidden True)
+        "loud"   -> Just (set winSilent False)
+        "silent" -> Just (set winSilent True)
+        _        -> Nothing
+
+tabSetWindow :: Bool {- ^ reversed -} -> ClientCommand String
+tabSetWindow isReversed st _ =
+  simpleTabCompletion plainWordCompleteMode [] completions isReversed st
+  where
+    completions = ["hide", "show", "loud", "silent"] :: [Text]
+
 -- | Implementation of @/grep@
 cmdGrep :: ClientCommand String
 cmdGrep st str
@@ -392,8 +461,8 @@ cmdClear st args =
 
     clearFocus1 focus st' = focusEffect (windowEffect st')
       where
-        windowEffect = set (clientWindows . at focus)
-                           (if isActive then Just emptyWindow else Nothing)
+        windowEffect = over (clientWindows . at focus)
+                           (if isActive then fmap windowClear else const Nothing)
 
         focusEffect
           | noChangeNeeded    = id
@@ -499,10 +568,8 @@ cmdDump st fp =
 
   where
     focus = view clientFocus st
-    msgs  = preview (clientWindows . ix focus . winMessages) st
-    outputLines =
-      case msgs of
-        Nothing  -> []
-        Just wls -> convert [] wls
-    convert acc Nil = acc
-    convert acc (wl :- wls) = convert (views wlFullImage imageText wl : acc) wls
+
+    outputLines
+      = reverse
+      $ clientFilter st id
+      $ toListOf (clientWindows . ix focus . winMessages . each . wlText) st
