@@ -4,6 +4,7 @@
 {-# LANGUAGE TemplateHaskell   #-}
 {-# LANGUAGE ApplicativeDo     #-}
 {-# LANGUAGE RankNTypes        #-}
+{-# LANGUAGE LambdaCase        #-}
 
 {-|
 Module      : Client.Configuration
@@ -21,6 +22,7 @@ module Client.Configuration
     Configuration(..)
   , ConfigurationFailure(..)
   , LayoutMode(..)
+  , EditMode(..)
   , PaddingMode(..)
   , ExtensionConfiguration(..)
 
@@ -30,10 +32,10 @@ module Client.Configuration
   , configPalette
   , configWindowNames
   , configNickPadding
-  , configDownloadDir
   , configMacros
   , configExtensions
   , configExtraHighlights
+  , configNeverHighlights
   , configUrlOpener
   , configIgnores
   , configActivityBar
@@ -43,6 +45,7 @@ module Client.Configuration
   , configLayout
   , configShowPing
   , configJumpModifier
+  , configDigraphs
 
   , extensionPath
   , extensionRtldFlags
@@ -52,7 +55,7 @@ module Client.Configuration
   , loadConfiguration
 
   -- * Resolving paths
-  , getNewConfigPath
+  , getConfigPath
 
   -- * Specification
   , configurationSpec
@@ -61,6 +64,10 @@ module Client.Configuration
   , FilePathContext
   , newFilePathContext
   , resolveFilePath
+
+  -- * Url opener configuration
+  , UrlOpener(..)
+  , UrlArgument(..)
   ) where
 
 import           Client.Commands.Interpolation
@@ -71,30 +78,29 @@ import           Client.Configuration.ServerSettings
 import           Client.EventLoop.Actions
 import           Client.Image.Palette
 import           Config
+import           Config.Macro
 import           Config.Schema
 import           Control.Exception
-import           Control.Monad                       (unless)
 import           Control.Lens                        hiding (List)
-import           Data.Foldable                       (toList)
-import           Data.Functor.Alt                    ((<!>))
+import           Data.Foldable                       (foldl', toList)
 import           Data.HashMap.Strict                 (HashMap)
 import qualified Data.HashMap.Strict                 as HashMap
-import           Data.HashSet                        (HashSet)
-import qualified Data.HashSet                        as HashSet
 import qualified Data.List.NonEmpty                  as NonEmpty
+import           Data.Map                            (Map)
+import qualified Data.Map                            as Map
 import           Data.Maybe
 import           Data.Monoid                         (Endo(..))
 import           Data.Text                           (Text)
 import qualified Data.Text                           as Text
-import qualified Data.Text.IO                        as Text
 import qualified Data.Vector                         as Vector
+import           Digraphs (Digraph(..))
 import           Graphics.Vty.Input.Events (Modifier(..), Key(..))
 import           Graphics.Vty.Attributes             (Attr)
 import           Irc.Identifier                      (Identifier)
 import           System.Directory
 import           System.FilePath
-import           System.IO.Error
 import           System.Posix.DynamicLinker          (RTLDFlags(..))
+import           System.IO.Error
 
 -- | Top-level client configuration information. When connecting to a
 -- server configuration from '_configServers' is used where possible,
@@ -104,12 +110,12 @@ data Configuration = Configuration
   , _configServers         :: (HashMap Text ServerSettings) -- ^ Host-specific settings
   , _configPalette         :: Palette -- ^ User-customized color palette
   , _configWindowNames     :: Text -- ^ Names of windows, used when alt-jumping)
-  , _configExtraHighlights :: HashSet Identifier -- ^ Extra highlight nicks/terms
+  , _configExtraHighlights :: [Identifier] -- ^ Extra highlight nicks/terms
+  , _configNeverHighlights :: [Identifier] -- ^ Never highlight nicks/terms
   , _configNickPadding     :: PaddingMode -- ^ Padding of nicks in messages
-  , _configDownloadDir     :: FilePath -- ^ Directory for downloads, default to HOME
   , _configMacros          :: Recognizer Macro -- ^ command macros
   , _configExtensions      :: [ExtensionConfiguration] -- ^ extensions to load
-  , _configUrlOpener       :: Maybe FilePath -- ^ paths to url opening executable
+  , _configUrlOpener       :: Maybe UrlOpener -- ^ paths to url opening executable
   , _configIgnores         :: [Text] -- ^ initial ignore mask list
   , _configActivityBar     :: Bool -- ^ initially visibility of the activity bar
   , _configBellOnMention   :: Bool -- ^ notify terminal on mention
@@ -118,9 +124,16 @@ data Configuration = Configuration
   , _configLayout          :: LayoutMode -- ^ Default layout on startup
   , _configShowPing        :: Bool -- ^ visibility of ping time
   , _configJumpModifier    :: [Modifier] -- ^ Modifier used for jumping windows
+  , _configDigraphs        :: Map Digraph Text -- ^ Extra digraphs
   }
   deriving Show
 
+data UrlOpener = UrlOpener FilePath [UrlArgument]
+  deriving Show
+
+data UrlArgument = UrlArgLiteral String | UrlArgUrl
+  deriving Show
+
 -- | Setting for how to pad the message prefix.
 data PaddingMode
   = LeftPadding  !Int -- ^ Whitespace add to the left side of chat prefix
@@ -135,6 +148,10 @@ data LayoutMode
   | TwoColumn
   deriving Show
 
+data EditMode
+  = SingleLineEditor
+  | MultiLineEditor
+  deriving Show
 
 -- | Failure cases when loading a configuration file.
 data ConfigurationFailure
@@ -170,59 +187,12 @@ makeLenses ''ExtensionConfiguration
 defaultWindowNames :: Text
 defaultWindowNames = "1234567890qwertyuiop!@#$%^&*()QWERTYUIOP"
 
--- | Uses 'getAppUserDataDirectory' to find @~/.glirc/config@
-getOldConfigPath :: IO FilePath
-getOldConfigPath =
-  do dir <- getAppUserDataDirectory "glirc"
-     return (dir </> "config")
-
 -- | Uses 'getXdgDirectory' 'XdgConfig' to find @~/.config/glirc/config@
-getNewConfigPath :: IO FilePath
-getNewConfigPath =
+getConfigPath :: IO FilePath
+getConfigPath =
   do dir <- getXdgDirectory XdgConfig "glirc"
      return (dir </> "config")
 
--- | Empty configuration file used when no path is specified
--- and the configuration file is missing.
-emptyConfigFile :: Text
-emptyConfigFile = "{}\n"
-
--- | Attempt to read a file using the given handler when
--- a file does not exist. On failure a 'ConfigurationReadFailed'
--- exception is throw.
-readFileCatchNotFound ::
-  FilePath {- ^ file to read -} ->
-  (IOError -> IO (FilePath, Text)) {- ^ error handler for not found case -} ->
-  IO (FilePath, Text)
-readFileCatchNotFound path onNotFound =
-  do res <- try (Text.readFile path)
-     case res of
-       Left e | isDoesNotExistError e -> onNotFound e
-              | otherwise -> throwIO (ConfigurationReadFailed (show e))
-       Right txt -> return (path, txt)
-
--- | Either read a configuration file from one of the default
--- locations, in which case no configuration found is equivalent
--- to an empty configuration, or from the specified file where
--- no configuration found is an error.
-readConfigurationFile ::
-  Maybe FilePath {- ^ just file or use default search paths -} ->
-  IO (FilePath, Text)
-readConfigurationFile mbPath =
-  case mbPath of
-
-    Just path ->
-      readFileCatchNotFound path $ \e ->
-        throwIO (ConfigurationReadFailed (show e))
-
-    Nothing ->
-      do newPath <- getNewConfigPath
-         readFileCatchNotFound newPath $ \_ ->
-           do oldPath <- getOldConfigPath
-              readFileCatchNotFound oldPath $ \_ ->
-                return ("", emptyConfigFile)
-
-
 -- | Load the configuration file defaulting to @~/.glirc/config@.
 --
 -- Given configuration path is optional and actual path used will
@@ -231,61 +201,61 @@ loadConfiguration ::
   Maybe FilePath {- ^ path to configuration file -} ->
   IO (Either ConfigurationFailure (FilePath, Configuration))
 loadConfiguration mbPath = try $
-  do (path,txt) <- readConfigurationFile mbPath
-     home <- getHomeDirectory
+  do path <- case mbPath of
+               Nothing -> getConfigPath
+               Just p  -> pure p
 
-     rawcfg <-
-       case parse txt of
-         Left e -> throwIO (ConfigurationParseFailed path (displayException e))
-         Right rawcfg -> return rawcfg
+     ctx <- newFilePathContext path
+
+     let toPath txt _ = pure (resolveFilePath ctx (Text.unpack txt))
+     rawcfg <- loadFileWithMacros toPath path
+        `catches`
+        [Handler $ \e -> case e of
+           LoadFileParseError fp pe -> throwIO (ConfigurationParseFailed fp (displayException pe))
+           LoadFileMacroError (UndeclaredVariable a var) -> badMacro a ("undeclared variable: " ++ Text.unpack var)
+           LoadFileMacroError (BadSplice a)              -> badMacro a "bad @splice"
+           LoadFileMacroError (BadLoad a)                -> badMacro a "bad @load"
+           LoadFileMacroError (UnknownDirective a dir)   -> badMacro a ("unknown directive: @" ++ Text.unpack dir)
+        ,Handler $ \e ->
+            if isDoesNotExistError e &&
+               isNothing mbPath &&
+               ioeGetFileName e == Just path
+            then pure (Sections (FilePosition path (Position 0 0 0)) [])
+            else throwIO (ConfigurationReadFailed (displayException e))
+        ]
 
      case loadValue configurationSpec rawcfg of
        Left e -> throwIO
                $ ConfigurationMalformed path
                $ displayException e
-       Right cfg ->
-         do cfg' <- resolvePaths path (cfg defaultServerSettings home)
-                    >>= validateDirectories path
-            return (path, cfg')
+       Right cfg -> return (path, resolvePaths ctx (cfg defaultServerSettings))
 
+badMacro :: FilePosition -> String -> IO a
+badMacro (FilePosition path posn) msg =
+  throwIO $ ConfigurationMalformed path
+          $ "line "    ++ show (posLine   posn) ++
+            " column " ++ show (posColumn posn) ++
+            ": "       ++ msg
 
 -- | Resolve all the potentially relative file paths in the configuration file
-resolvePaths :: FilePath -> Configuration -> IO Configuration
-resolvePaths file cfg =
-  do res <- resolveFilePath <$> newFilePathContext file
-     let resolveServerFilePaths = over (ssTlsClientCert . mapped) res
-                                . over (ssTlsClientKey  . mapped) res
-                                . over (ssTlsServerCert . mapped) res
-                                . over (ssSaslMechanism . mapped . _SaslEcdsa . _3) res
-                                . over (ssLogDir        . mapped) res
-     return $! over (configExtensions . mapped . extensionPath) res
-             . over (configServers    . mapped) resolveServerFilePaths
-             . over configDownloadDir res
-             $ cfg
-
--- | Check if the `download-dir` is actually a directory and writeable,
---   throw a ConfigurationMalformed exception if it isn't.
-validateDirectories :: FilePath -> Configuration -> IO Configuration
-validateDirectories cfgPath cfg =
-  do isDir       <- doesDirectoryExist downloadPath
-     unless isDir $ throwIO (ConfigurationMalformed cfgPath noDirMsg)
-     isWriteable <- writable <$> getPermissions downloadPath
-     unless isWriteable
-       $ throwIO (ConfigurationMalformed cfgPath noWriteableMsg)
-     return cfg
-  where
-    downloadPath = view configDownloadDir cfg
-    noDirMsg = "The download-dir section doesn't point to a directory."
-    noWriteableMsg = "The download-dir doesn't point to a writeable directory."
+resolvePaths :: FilePathContext -> Configuration -> Configuration
+resolvePaths ctx =
+  let res = resolveFilePath ctx
+      resolveServerFilePaths = over (ssTlsClientCert . mapped) res
+                             . over (ssTlsClientKey  . mapped) res
+                             . over (ssTlsServerCert . mapped) res
+                             . over (ssSaslMechanism . mapped . _SaslEcdsa . _3) res
+                             . over (ssLogDir        . mapped) res
+  in over (configExtensions . mapped . extensionPath) res
+   . over (configServers    . mapped) resolveServerFilePaths
 
 configurationSpec ::
-  ValueSpec (ServerSettings -> FilePath -> Configuration)
+  ValueSpec (ServerSettings -> Configuration)
 configurationSpec = sectionsSpec "config-file" $
 
   do let sec' def name spec info = fromMaybe def <$> optSection' name spec info
-         identifierSetSpec       = HashSet.fromList <$> listSpec identifierSpec
 
-     ssDefUpdate            <- sec' id "defaults" serverSpec
+     ssDefUpdate            <- sec' (Nothing,id) "defaults" serverSpec
                                "Default values for use across all server configurations"
      ssUpdates              <- sec' [] "servers" (listSpec serverSpec)
                                "Configuration parameters for IRC servers"
@@ -299,10 +269,12 @@ configurationSpec = sectionsSpec "config-file" $
                                "Programmable macro commands"
      _configExtensions      <- sec' [] "extensions" (listSpec extensionSpec)
                                "extension libraries to load at startup"
-     _configUrlOpener       <- optSection' "url-opener" stringSpec
+     _configUrlOpener       <- optSection' "url-opener" urlOpenerSpec
                                "External command used by /url command"
-     _configExtraHighlights <- sec' mempty "extra-highlights" identifierSetSpec
+     _configExtraHighlights <- sec' mempty "extra-highlights" (listSpec identifierSpec)
                                "Extra words to highlight in chat messages"
+     _configNeverHighlights <- sec' mempty "never-highlights" (listSpec identifierSpec)
+                               "Words to avoid highlighting in chat messages"
      _configNickPadding     <- sec' NoPadding "nick-padding" nickPaddingSpec
                                "Amount of space to reserve for nicknames in chat messages"
      _configIgnores         <- sec' [] "ignores" anySpec
@@ -320,13 +292,12 @@ configurationSpec = sectionsSpec "config-file" $
                                "Initial setting for window layout"
      _configShowPing        <- sec' True "show-ping" yesOrNoSpec
                                "Initial setting for visibility of ping times"
-     maybeDownloadDir       <- optSection' "download-dir" stringSpec
-                               "Path to DCC download directoy. Defaults to home directory."
-     return (\def home ->
-             let _configDefaults = ssDefUpdate def
+     _configDigraphs        <- sec' mempty "extra-digraphs" (Map.fromList <$> listSpec digraphSpec)
+                               "Extra digraphs"
+     return (\def ->
+             let _configDefaults = snd ssDefUpdate def
                  _configServers  = buildServerMap _configDefaults ssUpdates
                  _configKeyMap   = foldl (\acc f -> f acc) initialKeyMap bindings
-                 _configDownloadDir = fromMaybe home maybeDownloadDir
              in Configuration{..})
 
 -- | The default nick padding side if padding is going to be used
@@ -496,17 +467,55 @@ rtldFlagSpec = namedSpec "rtld-flag"
            <!> RTLD_NOW    <$ atomSpec "now"
            <!> RTLD_LAZY   <$ atomSpec "lazy"
 
+urlOpenerSpec :: ValueSpec UrlOpener
+urlOpenerSpec = simpleCase <!> complexCase
+  where
+    simpleCase =
+      do path <- stringSpec
+         pure (UrlOpener path [UrlArgUrl])
+
+    complexCase = sectionsSpec "url-opener" $
+      do path <- reqSection' "path" stringSpec "Executable"
+         args <- reqSection' "args" (listSpec argSpec) "Arguments"
+         pure (UrlOpener path args)
+
+    argSpec = UrlArgUrl     <$  atomSpec "url"
+          <!> UrlArgLiteral <$> stringSpec
+
+digraphSpec :: ValueSpec (Digraph, Text)
+digraphSpec =
+  sectionsSpec "digraph" $
+   do ~(x,y) <- reqSection' "input" twoChars "digraph key"
+      val    <- reqSection' "output" textSpec "replacement text"
+      pure (Digraph x y, val)
+  where
+    twoChars = customSpec "two" stringSpec $ \case
+        [x,y] -> Right (x,y)
+        _     -> Left "exactly two characters required"
+
 buildServerMap ::
   ServerSettings {- ^ defaults -} ->
-  [ServerSettings -> ServerSettings] ->
+  [(Maybe Text, ServerSettings -> ServerSettings)] ->
   HashMap Text ServerSettings
-buildServerMap def ups =
-  HashMap.fromList [ (serverSettingName ss, ss) | up <- ups, let ss = up def ]
+buildServerMap def ups = go HashMap.empty def Nothing
   where
     serverSettingName ss =
       fromMaybe (views ssHostName Text.pack ss)
                 (view ssName ss)
 
+    raw = HashMap.fromListWith (++) [(mbExtName, [up]) | (mbExtName, up) <- ups]
+
+    go acc prev prevName = foldl' (add prev) acc nexts
+      where
+        nexts = HashMap.findWithDefault [] prevName raw
+
+    add prev acc f = go acc' ss (Just me)
+      where
+        ss = f prev
+        me = serverSettingName ss
+        acc'
+          | HashMap.member me acc = acc
+          | otherwise             = HashMap.insert me ss acc
 
 data FilePathContext = FilePathContext { fpBase, fpHome :: FilePath }
 
