@@ -1,4 +1,4 @@
-{-# LANGUAGE ApplicativeDo, TemplateHaskell, OverloadedStrings #-}
+{-# LANGUAGE ApplicativeDo, TemplateHaskell, OverloadedStrings, RecordWildCards, BlockArguments #-}
 
 {-|
 Module      : Client.Configuration.ServerSettings
@@ -29,11 +29,13 @@ module Client.Configuration.ServerSettings
   , ssHostName
   , ssPort
   , ssTls
+  , ssTlsVerify
   , ssTlsClientCert
   , ssTlsClientKey
   , ssTlsClientKeyPassword
   , ssTlsServerCert
   , ssTlsCiphers
+  , ssTls13Ciphers
   , ssConnectCmds
   , ssSocksHost
   , ssSocksPort
@@ -59,6 +61,7 @@ module Client.Configuration.ServerSettings
   , _SaslExternal
   , _SaslEcdsa
   , _SaslPlain
+  , _SaslScram
 
   -- * Secrets
   , Secret(..)
@@ -71,12 +74,14 @@ module Client.Configuration.ServerSettings
   -- * TLS settings
   , UseTls(..)
   , Fingerprint(..)
+  , TlsMode(..)
 
   -- * Regex wrapper
   , KnownRegex(..)
   , getRegex
   ) where
 
+import           Client.Authentication.Scram (ScramDigest(..))
 import           Client.Commands.Interpolation
 import           Client.Commands.WordCompletion
 import           Client.Configuration.Macros (macroCommandSpec)
@@ -85,13 +90,13 @@ import           Control.Exception (Exception, displayException, throwIO, try)
 import           Control.Lens
 import           Control.Monad ((>=>))
 import qualified Data.ByteString as B
-import           Data.Functor.Alt                    ((<!>))
-import           Data.List.NonEmpty (NonEmpty((:|)))
 import           Data.ByteString (ByteString)
-import           Data.Monoid
-import           Data.Text (Text)
+import           Data.List.NonEmpty (NonEmpty((:|)))
 import qualified Data.List.NonEmpty as NonEmpty
 import           Data.List.Split (chunksOf, splitOn)
+import           Data.Maybe (fromMaybe)
+import           Data.Monoid
+import           Data.Text (Text)
 import qualified Data.Text as Text
 import           Irc.Identifier (Identifier, mkId)
 import           Network.Socket (HostName, PortNumber)
@@ -100,6 +105,7 @@ import qualified System.Exit as Exit
 import qualified System.Process as Process
 import           Text.Regex.TDFA
 import           Text.Regex.TDFA.Text (compile)
+import           Hookup (TlsVerify(..))
 
 -- | Static server-level settings
 data ServerSettings = ServerSettings
@@ -110,12 +116,17 @@ data ServerSettings = ServerSettings
   , _ssSaslMechanism    :: !(Maybe SaslMechanism) -- ^ SASL mechanism
   , _ssHostName         :: !HostName -- ^ server hostname
   , _ssPort             :: !(Maybe PortNumber) -- ^ server port
-  , _ssTls              :: !UseTls -- ^ use TLS to connect
+  , _ssTls              :: !TlsMode -- ^ use TLS to connect
+  , _ssTlsVerify        :: !TlsVerify -- ^ verify TLS hostname
   , _ssTlsClientCert    :: !(Maybe FilePath) -- ^ path to client TLS certificate
   , _ssTlsClientKey     :: !(Maybe FilePath) -- ^ path to client TLS key
   , _ssTlsClientKeyPassword :: !(Maybe Secret) -- ^ client key PEM password
   , _ssTlsServerCert    :: !(Maybe FilePath) -- ^ additional CA certificates for validating server
   , _ssTlsCiphers       :: String            -- ^ OpenSSL cipher suite
+  , _ssTls13Ciphers     :: Maybe String      -- ^ OpenSSL TLS 1.3 cipher suite
+  , _ssTlsPubkeyFingerprint :: !(Maybe Fingerprint) -- ^ optional acceptable public key fingerprint
+  , _ssTlsCertFingerprint   :: !(Maybe Fingerprint) -- ^ optional acceptable certificate fingerprint
+  , _ssSts              :: !Bool -- ^ Honor STS policies when true
   , _ssConnectCmds      :: ![[ExpansionChunk]] -- ^ commands to execute upon successful connection
   , _ssSocksHost        :: !(Maybe HostName) -- ^ hostname of SOCKS proxy
   , _ssSocksPort        :: !PortNumber -- ^ port of SOCKS proxy
@@ -130,14 +141,14 @@ data ServerSettings = ServerSettings
   , _ssNickCompletion   :: WordCompletionMode -- ^ Nick completion mode for this server
   , _ssLogDir           :: Maybe FilePath -- ^ Directory to save logs of chat
   , _ssBindHostName     :: Maybe HostName -- ^ Local bind host
-  , _ssSts              :: !Bool -- ^ Honor STS policies when true
-  , _ssTlsPubkeyFingerprint :: !(Maybe Fingerprint) -- ^ optional acceptable public key fingerprint
-  , _ssTlsCertFingerprint   :: !(Maybe Fingerprint) -- ^ optional acceptable certificate fingerprint
   , _ssShowAccounts     :: !Bool -- ^ Render account names
   , _ssCapabilities     :: ![Text] -- ^ Extra capabilities to unconditionally request
   }
   deriving Show
 
+data TlsMode = TlsYes | TlsNo | TlsStart
+  deriving Show
+
 data Secret
   = SecretText Text    -- ^ Constant text
   | SecretCommand (NonEmpty Text) -- ^ Command to generate text
@@ -148,6 +159,8 @@ data SaslMechanism
   = SaslPlain    (Maybe Text) Text Secret -- ^ SASL PLAIN RFC4616 - authzid authcid password
   | SaslEcdsa    (Maybe Text) Text FilePath -- ^ SASL NIST - https://github.com/kaniini/ecdsatool - authzid keypath
   | SaslExternal (Maybe Text)      -- ^ SASL EXTERNAL RFC4422 - authzid
+  | SaslScram    ScramDigest (Maybe Text) Text Secret -- ^ SASL SCRAM-SHA-256 RFC7677 - authzid authcid password
+  | SaslEcdh     (Maybe Text) Text Secret -- ^ SASL ECDH-X25519-CHALLENGE - authzid authcid private-key
   deriving Show
 
 -- | Regular expression matched with original source to help with debugging.
@@ -190,12 +203,17 @@ defaultServerSettings =
        , _ssSaslMechanism = Nothing
        , _ssHostName      = ""
        , _ssPort          = Nothing
-       , _ssTls           = UseInsecure
+       , _ssTls           = TlsNo
+       , _ssTlsVerify     = VerifyDefault
        , _ssTlsClientCert = Nothing
        , _ssTlsClientKey  = Nothing
        , _ssTlsClientKeyPassword = Nothing
        , _ssTlsServerCert = Nothing
        , _ssTlsCiphers    = "HIGH"
+       , _ssTls13Ciphers  = Nothing
+       , _ssTlsPubkeyFingerprint = Nothing
+       , _ssTlsCertFingerprint   = Nothing
+       , _ssSts              = True
        , _ssConnectCmds   = []
        , _ssSocksHost     = Nothing
        , _ssSocksPort     = 1080
@@ -210,16 +228,15 @@ defaultServerSettings =
        , _ssNickCompletion   = defaultNickWordCompleteMode
        , _ssLogDir           = Nothing
        , _ssBindHostName     = Nothing
-       , _ssSts              = True
-       , _ssTlsPubkeyFingerprint = Nothing
-       , _ssTlsCertFingerprint   = Nothing
        , _ssShowAccounts     = False
        , _ssCapabilities     = []
        }
 
-serverSpec :: ValueSpec (ServerSettings -> ServerSettings)
+serverSpec :: ValueSpec (Maybe Text, ServerSettings -> ServerSettings)
 serverSpec = sectionsSpec "server-settings" $
-  composeMaybe <$> sequenceA settings
+  do mbExt <- optSection "extends" "name of a server to use for defaults"
+     upd <- composeMaybe <$> sequenceA settings
+     pure (mbExt, upd)
   where
 
     composeMaybe :: [Maybe (a -> a)] -> a -> a
@@ -260,24 +277,30 @@ serverSpec = sectionsSpec "server-settings" $
       , opt "sasl" ssSaslMechanism saslMechanismSpec
         "SASL settings"
 
-      , req "tls" ssTls useTlsSpec
-        "Set to `yes` to enable secure connect. Set to `yes-insecure` to disable certificate checking."
+      , req "tls" ssTls tlsModeSpec
+        "Use TLS to connect (default no)"
 
-      , opt "tls-client-cert" ssTlsClientCert stringSpec
+      , req "tls-verify" ssTlsVerify tlsVerifySpec
+        "Enable server certificate hostname verification (default yes, string to override hostname)"
+
+      , opt "tls-client-cert" ssTlsClientCert filepathSpec
         "Path to TLS client certificate"
 
-      , opt "tls-client-key" ssTlsClientKey stringSpec
+      , opt "tls-client-key" ssTlsClientKey filepathSpec
         "Path to TLS client key"
 
       , opt "tls-client-key-password" ssTlsClientKeyPassword anySpec
         "Password for decrypting TLS client key PEM file"
 
-      , opt "tls-server-cert" ssTlsServerCert stringSpec
+      , opt "tls-server-cert" ssTlsServerCert filepathSpec
         "Path to CA certificate bundle"
 
       , req "tls-ciphers" ssTlsCiphers stringSpec
         "OpenSSL cipher specification. Default to \"HIGH\""
 
+      , opt "tls-1.3-ciphers" ssTls13Ciphers stringSpec
+        "OpenSSL TLS 1.3 cipher specification."
+
       , opt "socks-host" ssSocksHost stringSpec
         "Hostname of SOCKS5 proxy server"
 
@@ -311,7 +334,7 @@ serverSpec = sectionsSpec "server-settings" $
       , req "nick-completion" ssNickCompletion nickCompletionSpec
         "Behavior for nickname completion with TAB"
 
-      , opt "log-dir" ssLogDir stringSpec
+      , opt "log-dir" ssLogDir filepathSpec
         "Path to log file directory for this server"
 
       , opt "bind-hostname" ssBindHostName stringSpec
@@ -333,8 +356,20 @@ serverSpec = sectionsSpec "server-settings" $
         "Extra capabilities to unconditionally request from the server"
       ]
 
+tlsModeSpec :: ValueSpec TlsMode
+tlsModeSpec =
+  TlsYes   <$ atomSpec "yes"      <!>
+  TlsNo    <$ atomSpec "no"       <!>
+  TlsStart <$ atomSpec "starttls"
+
+tlsVerifySpec :: ValueSpec TlsVerify
+tlsVerifySpec =
+  VerifyDefault  <$ atomSpec "yes"      <!>
+  VerifyNone     <$ atomSpec "no"       <!>
+  VerifyHostname <$> stringSpec
+
 saslMechanismSpec :: ValueSpec SaslMechanism
-saslMechanismSpec = plain <!> external <!> ecdsa
+saslMechanismSpec = plain <!> external <!> ecdsa <!> scram <!> ecdh
   where
     mech m   = reqSection' "mechanism" (atomSpec m) "Mechanism"
     authzid  = optSection "authzid" "Authorization identity"
@@ -353,7 +388,36 @@ saslMechanismSpec = plain <!> external <!> ecdsa
       sectionsSpec "sasl-ecdsa-nist256p-challenge-mech" $
       SaslEcdsa <$ mech "ecdsa-nist256p-challenge" <*>
       authzid <*> username <*>
-      reqSection' "private-key" stringSpec "Private key file"
+      reqSection' "private-key" filepathSpec "Private key file"
+
+    scramDigest =
+      fromMaybe ScramDigestSha2_256 <$>
+      optSection' "digest" scramDigests "Underlying digest function"
+
+    scramDigests =
+      ScramDigestSha1     <$ atomSpec "sha1" <!>
+      ScramDigestSha2_256 <$ atomSpec "sha2-256" <!>
+      ScramDigestSha2_512 <$ atomSpec "sha2-512"
+
+    scram =
+      sectionsSpec "sasl-scram" $
+      SaslScram <$ mech "scram" <*>
+      scramDigest <*>
+      authzid <*> username <*> reqSection "password" "Password"
+
+    ecdh =
+      sectionsSpec "sasl-ecdh-x25519-challenge" $
+      SaslEcdh <$ mech "ecdh-x25519-challenge" <*>
+      authzid <*> username <*> reqSection "private-key" "Private Key"
+
+
+
+
+filepathSpec :: ValueSpec FilePath
+filepathSpec = customSpec "path" stringSpec $ \str ->
+  if null str
+  then Left "empty path"
+  else Right str
 
 hookSpec :: ValueSpec HookConfig
 hookSpec =
@@ -393,17 +457,24 @@ fingerprintSpec =
 nicksSpec :: ValueSpec (NonEmpty Text)
 nicksSpec = oneOrNonemptySpec anySpec
 
-useTlsSpec :: ValueSpec UseTls
-useTlsSpec =
-      UseTls         <$ atomSpec "yes"
-  <!> UseInsecureTls <$ atomSpec "yes-insecure"
-  <!> UseInsecure    <$ atomSpec "no"
-
-
 nickCompletionSpec :: ValueSpec WordCompletionMode
 nickCompletionSpec =
       defaultNickWordCompleteMode <$ atomSpec "default"
   <!> slackNickWordCompleteMode   <$ atomSpec "slack"
+  <!> customNickCompletion
+
+customNickCompletion :: ValueSpec WordCompletionMode
+customNickCompletion =
+  sectionsSpec "nick-completion" $
+  do wcmStartPrefix  <- fromMaybe "" <$> optSection' "start-prefix" stringSpec
+                        "Prefix for nickname with when completing at start of line."
+     wcmStartSuffix  <- fromMaybe "" <$> optSection' "start-suffix" stringSpec
+                        "Suffix for nickname with when completing at start of line."
+     wcmMiddlePrefix <- fromMaybe "" <$> optSection' "middle-prefix" stringSpec
+                        "Prefix for nickname with when completing in middle of line."
+     wcmMiddleSuffix <- fromMaybe "" <$> optSection' "middle-suffix" stringSpec
+                        "Suffix for nickname with when completing in middle of line."
+     pure WordCompletionMode{..}
 
 
 identifierSpec :: ValueSpec Identifier
@@ -430,7 +501,9 @@ loadSecrets :: ServerSettings -> IO ServerSettings
 loadSecrets =
   traverseOf (ssPassword             . _Just                  ) (loadSecret "server password") >=>
   traverseOf (ssSaslMechanism        . _Just . _SaslPlain . _3) (loadSecret "SASL password") >=>
-  traverseOf (ssTlsClientKeyPassword . _Just                  ) (loadSecret "TLS key password")
+  traverseOf (ssTlsClientKeyPassword . _Just                  ) (loadSecret "TLS key password") >=>
+  traverseOf (ssSaslMechanism        . _Just . _SaslScram . _4) (loadSecret "SASL password") >=>
+  traverseOf (ssSaslMechanism        . _Just . _SaslEcdh  . _3) (loadSecret "SASL private key")
 
 -- | Run a command if found and replace it with the first line of stdout result.
 loadSecret :: String -> Secret -> IO Secret
