@@ -14,11 +14,12 @@ events to the correct module. It renders the user interface once per event.
 module Client.EventLoop
   ( eventLoop
   , updateTerminalSize
+  , ClientEvent(..)
   ) where
 
-import           Client.CApi (popTimer)
+import           Client.CApi (ThreadEntry, popTimer)
 import           Client.Commands
-import           Client.Configuration (configJumpModifier, configKeyMap, configWindowNames, configDownloadDir)
+import           Client.Configuration (configJumpModifier, configKeyMap, configWindowNames, configDigraphs)
 import           Client.Configuration.ServerSettings
 import           Client.EventLoop.Actions
 import           Client.EventLoop.Errors (exceptionToLines)
@@ -26,12 +27,12 @@ import           Client.EventLoop.Network (clientResponse)
 import           Client.Hook
 import           Client.Image
 import           Client.Image.Layout (scrollAmount)
+import           Client.Image.StatusLine (clientTitle)
 import           Client.Log
 import           Client.Message
 import           Client.Network.Async
 import           Client.State
-import           Client.State.DCC
-import qualified Client.State.EditBox     as Edit
+import qualified Client.State.EditBox as Edit
 import           Client.State.Extensions
 import           Client.State.Focus
 import           Client.State.Network
@@ -40,6 +41,7 @@ import           Control.Exception
 import           Control.Lens
 import           Control.Monad
 import           Data.ByteString (ByteString)
+import           Data.Char (isSpace)
 import           Data.Foldable
 import           Data.Traversable
 import           Data.List
@@ -54,7 +56,9 @@ import qualified Data.Text.Encoding.Error as Text
 import           Data.Time
 import           GHC.IO.Exception (IOErrorType(..), ioe_type)
 import           Graphics.Vty
+import           Graphics.Vty.Input.Events
 import           Irc.Message
+import           Irc.Codes
 import           Irc.RawIrcMsg
 import           LensUtils
 import           Hookup (ConnectionFailure(..))
@@ -62,11 +66,11 @@ import           Hookup (ConnectionFailure(..))
 
 -- | Sum of the five possible event types the event loop handles
 data ClientEvent
-  = VtyEvent Event                        -- ^ Key presses and resizing
+  = VtyEvent InternalEvent -- ^ Key presses and resizing
   | NetworkEvents (NonEmpty (Text, NetworkEvent)) -- ^ Incoming network events
   | TimerEvent Text TimedAction      -- ^ Timed action and the applicable network
   | ExtTimerEvent Int                     -- ^ extension ID
-  | DCCUpdate DCCUpdate                   -- ^ Event on any transfer
+  | ThreadEvent Int ThreadEntry
 
 
 -- | Block waiting for the next 'ClientEvent'. This function will compute
@@ -77,7 +81,7 @@ getEvent ::
   IO ClientEvent
 getEvent vty st =
   do timer <- prepareTimer
-     atomically (asum [timer, vtyEvent, networkEvents, dccUpdate])
+     atomically (asum [timer, vtyEvent, networkEvents, threadJoin])
   where
     vtyEvent = VtyEvent <$> readTChan (_eventChannel (inputIface vty))
 
@@ -100,7 +104,9 @@ getEvent vty st =
                          unless ready retry
                          return event
 
-    dccUpdate = DCCUpdate <$> readTChan (view clientDCCUpdates st)
+    threadJoin =
+      do (i,r) <- readTQueue (view clientThreadJoins st)
+         pure (ThreadEvent i r)
 
 -- | Compute the earliest scheduled timed action for the client
 earliestEvent :: ClientState -> Maybe (UTCTime, ClientEvent)
@@ -133,19 +139,22 @@ eventLoop vty st =
 
      let (pic, st') = clientPicture (clientTick st)
      update vty pic
+     setWindowTitle vty (clientTitle st)
 
      event <- getEvent vty st'
      case event of
        ExtTimerEvent i ->
          eventLoop vty =<< clientExtTimer i st'
+       ThreadEvent i result ->
+         eventLoop vty =<< clientThreadJoin i result st'
        TimerEvent networkId action ->
          eventLoop vty =<< doTimerEvent networkId action st'
-       VtyEvent vtyEvent ->
+       VtyEvent (InputEvent vtyEvent) ->
          traverse_ (eventLoop vty) =<< doVtyEvent vty vtyEvent st'
+       VtyEvent ResumeAfterSignal ->
+         eventLoop vty =<< updateTerminalSize vty st
        NetworkEvents networkEvents ->
          eventLoop vty =<< foldM doNetworkEvent st' networkEvents
-       DCCUpdate upd ->
-         eventLoop vty =<< doDCCUpdate upd st'
 
 -- | Apply a single network event to the client state.
 doNetworkEvent :: ClientState -> (Text, NetworkEvent) -> IO ClientState
@@ -153,7 +162,8 @@ doNetworkEvent st (net, networkEvent) =
   case networkEvent of
     NetworkLine  time line -> doNetworkLine  net time line st
     NetworkError time ex   -> doNetworkError net time ex st
-    NetworkOpen  time msg  -> doNetworkOpen  net time msg st
+    NetworkOpen  time      -> doNetworkOpen  net time st
+    NetworkTLS   txts      -> doNetworkTLS   net txts st
     NetworkClose time      -> doNetworkClose net time st
 
 -- | Sound the terminal bell assuming that the @BEL@ control code
@@ -169,10 +179,9 @@ processLogEntries =
 doNetworkOpen ::
   Text        {- ^ network name -} ->
   ZonedTime   {- ^ event time   -} ->
-  [Text]      {- ^ rendered certificate -} ->
   ClientState {- ^ client state -} ->
   IO ClientState
-doNetworkOpen networkId time cert st =
+doNetworkOpen networkId time st =
   case view (clientConnections . at networkId) st of
     Nothing -> error "doNetworkOpen: Network missing"
     Just cs ->
@@ -182,10 +191,21 @@ doNetworkOpen networkId time cert st =
                      , _msgBody    = NormalBody "connection opened"
                      }
          let cs' = cs & csLastReceived .~ (Just $! zonedTimeToUTC time)
-                      & csCertificate  .~ cert
          return $! recordNetworkMessage msg
                  $ setStrict (clientConnections . ix networkId) cs' st
 
+-- | Update the TLS certificates for a connection
+doNetworkTLS ::
+  Text   {- ^ network name      -} ->
+  [Text] {- ^ certificate lines -} ->
+  ClientState ->
+  IO ClientState
+doNetworkTLS network cert st =
+  pure $! over (clientConnections . ix network) upd st
+  where
+    upd = set csCertificate cert
+        . set (csPingStatus . _PingConnecting . _3) NoRestriction
+
 -- | Respond to a network connection closing normally.
 doNetworkClose ::
   Text        {- ^ network name -} ->
@@ -231,7 +251,7 @@ reconnectLogicOnFailure ex cs st
   where
     computeRetryInfo =
       case view csPingStatus cs of
-        PingConnecting n tm                   -> pure (n+1, tm)
+        PingConnecting n tm _                 -> pure (n+1, tm)
         _ | Just tm <- view csLastReceived cs -> pure (1, Just tm)
           | otherwise                         -> do now <- getCurrentTime
                                                     pure (1, Just now)
@@ -240,7 +260,7 @@ reconnectLogicOnFailure ex cs st
 
     shouldReconnect =
       case view csPingStatus cs of
-        PingConnecting n _ | n == 0 || n > reconnectAttempts          -> False
+        PingConnecting n _ _ | n == 0 || n > reconnectAttempts        -> False
         _ | Just ConnectionFailure{}         <-      fromException ex -> True
           | Just HostnameResolutionFailure{} <-      fromException ex -> True
           | Just PingTimeout         <-              fromException ex -> True
@@ -263,9 +283,16 @@ doNetworkLine networkId time line st =
     Just cs ->
       let network = view csNetwork cs in
       case parseRawIrcMsg (asUtf8 line) of
+        _ | PingConnecting _ _ WaitTLSRestriction <- view csPingStatus cs ->
+          st <$ abortConnection StartTLSFailed (view csSocket cs)
+
+        Just raw
+          | PingConnecting _ _ StartTLSRestriction <- view csPingStatus cs ->
+          startTLSLine networkId cs st raw
+
         Nothing ->
           do let msg = Text.pack ("Malformed message: " ++ show line)
-             return $! recordError time (view csNetwork cs) msg st
+             return $! recordError time network msg st
 
         Just raw ->
           do (st1,passed) <- clientNotifyExtensions network raw st
@@ -284,26 +311,52 @@ doNetworkLine networkId time line st =
                Just irc ->
                  do -- state with message recorded
                     -- record messages *before* applying state changes
+                    -- so that quits, nick-changes, etc (TargetUser) will dispatch to
+                    -- the correct window
                     let st2 =
                           case viewHook irc of
-                            Nothing   -> st1 -- Message hidden
-                            Just irc' -> recordIrcMessage network target msg st1
+                            Nothing -> st1 -- Message hidden
+                            Just irc'
+                              | hideMessage irc' -> st1
+                              | otherwise -> recordIrcMessage network target msg st1
                               where
                                 myNick = view csNick cs
                                 target = msgTarget myNick irc
                                 msg = ClientMessage
-                                        { _msgTime    = time'
-                                        , _msgNetwork = network
-                                        , _msgBody    = IrcBody irc'
-                                        }
+                                      { _msgTime    = time'
+                                      , _msgNetwork = network
+                                      , _msgBody    = IrcBody irc'
+                                      }
 
-                    let (replies, dccUp, st3) =
+                    let (replies, st3) =
                           applyMessageToClientState time irc networkId cs st2
 
-                    traverse_ (atomically . writeTChan (view clientDCCUpdates st3)) dccUp
                     traverse_ (sendMsg cs) replies
                     clientResponse time' irc cs st3
 
+-- Highly restricted message handler for messages sent before STARTTLS
+-- has completed.
+startTLSLine :: Text -> NetworkState -> ClientState -> RawIrcMsg -> IO ClientState
+startTLSLine network cs st raw =
+  do now <- getZonedTime
+     let irc = cookIrcMsg raw
+         myNick = view csNick cs
+         target = msgTarget myNick irc
+         msg = ClientMessage
+             { _msgTime    = now
+             , _msgNetwork = network
+             , _msgBody    = IrcBody irc
+             }
+         st1 = recordIrcMessage network target msg st
+
+     case irc of
+       Notice{} -> pure st1
+       Reply _ RPL_STARTTLS _ ->
+         do upgrade (view csSocket cs)
+            pure (set ( clientConnections . ix network . csPingStatus
+                      . _PingConnecting . _3)
+                      WaitTLSRestriction st1)
+       _ -> st1 <$ abortConnection StartTLSFailed (view csSocket cs)
 
 -- | Find the ZNC provided server time
 computeEffectiveTime :: ZonedTime -> [TagEntry] -> ZonedTime
@@ -389,8 +442,8 @@ doAction vty action st =
     -- edits
     ActKillHome          -> changeEditor Edit.killHome
     ActKillEnd           -> changeEditor Edit.killEnd
-    ActKillWordBack      -> changeEditor (Edit.killWordBackward True)
-    ActKillWordForward   -> changeEditor (Edit.killWordForward True)
+    ActKillWordBack      -> changeEditor (Edit.killWordBackward isSpace True)
+    ActKillWordForward   -> changeEditor (Edit.killWordForward isSpace True)
     ActYank              -> changeEditor Edit.yank
     ActToggle            -> changeContent Edit.toggle
     ActDelete            -> changeContent Edit.delete
@@ -400,10 +453,12 @@ doAction vty action st =
     ActBold              -> changeEditor (Edit.insert '\^B')
     ActColor             -> changeEditor (Edit.insert '\^C')
     ActItalic            -> changeEditor (Edit.insert '\^]')
+    ActStrikethrough     -> changeEditor (Edit.insert '\^^')
     ActUnderline         -> changeEditor (Edit.insert '\^_')
     ActClearFormat       -> changeEditor (Edit.insert '\^O')
     ActReverseVideo      -> changeEditor (Edit.insert '\^V')
-    ActDigraph           -> mbChangeEditor Edit.insertDigraph
+    ActMonospace         -> changeEditor (Edit.insert '\^Q')
+    ActDigraph           -> mbChangeEditor (Edit.insertDigraph (view (clientConfig . configDigraphs) st))
     ActInsertEnter       -> changeEditor (Edit.insert '\^J')
 
     -- focus jumps
@@ -415,16 +470,20 @@ doAction vty action st =
     ActAdvanceNetwork    -> continue (advanceNetworkFocus st)
 
     ActReset             -> continue (changeSubfocus FocusMessages st)
-    ActOlderLine         -> changeEditor $ \ed      -> fromMaybe ed $ Edit.earlier ed
-    ActNewerLine         -> changeEditor $ \ed      -> fromMaybe ed $ Edit.later ed
+    ActOlderLine         -> changeEditor $ \ed -> fromMaybe ed $ Edit.earlier ed
+    ActNewerLine         -> changeEditor $ \ed -> fromMaybe ed $ Edit.later ed
     ActScrollUp          -> continue (scrollClient ( scrollAmount st) st)
     ActScrollDown        -> continue (scrollClient (-scrollAmount st) st)
+    ActScrollUpSmall     -> continue (scrollClient ( 3) st)
+    ActScrollDownSmall   -> continue (scrollClient (-3) st)
 
     ActTabCompleteBack   -> doCommandResult False =<< tabCompletion True  st
     ActTabComplete       -> doCommandResult False =<< tabCompletion False st
 
     ActInsert c          -> changeEditor (Edit.insert c)
-    ActEnter             -> doCommandResult True =<< executeInput st
+    ActEnter             -> if view clientEditLock st
+                            then when (view clientBell st) (beep vty) >> continue st
+                            else doCommandResult True =<< executeInput st
     ActRefresh           -> refresh vty >> continue st
     ActCommand cmd       -> do resp <- executeUserCommand Nothing (Text.unpack cmd) st
                                case resp of
@@ -472,24 +531,3 @@ doTimerEvent networkId action =
   traverseOf
     (clientConnection networkId)
     (applyTimedAction action)
-
-doDCCUpdate :: DCCUpdate -> ClientState -> IO ClientState
-doDCCUpdate upd st0 =
-  case upd of
-    PercentUpdate k newVal -> return $ set (commonLens k . dtProgress) newVal st0
-    SocketInterrupted k    -> reportKill k LostConnection
-    UserInterrupted k      -> reportKill k UserKilled
-    Finished k             -> reportKill k CorrectlyFinished
-    accept@(Accept k _ _) ->
-      do let dccState0 = view clientDCC st0
-             dccState1 = acceptUpdate accept dccState0
-             -- st1 = set clientDCC dccState1 st0
-
-             updChan = view clientDCCUpdates st0
-             mdir = view (clientConfig . configDownloadDir) st0
-
-         dccState2 <- supervisedDownload mdir k updChan dccState1
-         return $ set clientDCC dccState2 st0
-  where
-    reportKill k status = return $ over clientDCC (reportStopWithStatus k status) st0
-    commonLens k = clientDCC . dsTransfers . at k . _Just
