@@ -51,6 +51,8 @@ data Action
   | ActNewerLine
   | ActScrollUp
   | ActScrollDown
+  | ActScrollUpSmall
+  | ActScrollDownSmall
   | ActBackWord
   | ActForwardWord
 
@@ -65,7 +67,9 @@ data Action
   | ActColor
   | ActItalic
   | ActUnderline
+  | ActStrikethrough
   | ActReverseVideo
+  | ActMonospace
   | ActClearFormat
   | ActInsertEnter
   | ActDigraph
@@ -75,7 +79,7 @@ data Action
   | ActAdvanceNetwork
   | ActJumpToActivity
   | ActJumpPrevious
-  | ActJump Int
+  | ActJump Char
 
   | ActTabComplete
   | ActTabCompleteBack
@@ -133,9 +137,11 @@ actionInfos =
   ,("bold"              , (ActBold             , [ctrl (KChar 'b')]))
   ,("color"             , (ActColor            , [ctrl (KChar 'c')]))
   ,("italic"            , (ActItalic           , [ctrl (KChar ']')]))
+  ,("strikethrough"     , (ActStrikethrough    , [ctrl (KChar '^')]))
   ,("underline"         , (ActUnderline        , [ctrl (KChar '_')]))
   ,("clear-format"      , (ActClearFormat      , [ctrl (KChar 'o')]))
   ,("reverse-video"     , (ActReverseVideo     , [ctrl (KChar 'v')]))
+  ,("monospace"         , (ActMonospace        , [ctrl (KChar 'q')]))
 
   ,("insert-newline"    , (ActInsertEnter      , [meta KEnter]))
   ,("insert-digraph"    , (ActDigraph          , [meta (KChar 'k')]))
@@ -157,6 +163,8 @@ actionInfos =
   ,("down"              , (ActNewerLine        , [norm KDown]))
   ,("scroll-up"         , (ActScrollUp         , [norm KPageUp]))
   ,("scroll-down"       , (ActScrollDown       , [norm KPageDown]))
+  ,("scroll-up-small"   , (ActScrollUpSmall    , [meta KPageUp]))
+  ,("scroll-down-small" , (ActScrollDownSmall  , [meta KPageDown]))
   ,("enter"             , (ActEnter            , [norm KEnter]))
   ,("word-complete-back", (ActTabCompleteBack  , [norm KBackTab]))
   ,("word-complete"     , (ActTabComplete      , [norm (KChar '\t')]))
@@ -189,7 +197,7 @@ keyToAction ::
   Action     {- ^ action          -}
 keyToAction _ jumpMods names mods (KChar c)
   | normalizeModifiers jumpMods == normalizeModifiers mods
-  , Just i <- Text.findIndex (c==) names = ActJump i
+  , Text.singleton c `Text.isInfixOf` names = ActJump c
 keyToAction m _ _ modifier key =
   case m ^. keyMapLens modifier key of
     Just a -> a
@@ -229,6 +237,8 @@ initialKeyMap = KeyMap $
           , (KFun 3, ActCommand "toggle-activity-bar")
           , (KFun 4, ActCommand "toggle-metadata")
           , (KFun 5, ActCommand "toggle-layout")
+          , (KFun 6, ActCommand "toggle-editor")
+          , (KFun 7, ActCommand "toggle-edit-lock")
           ])
     :
 
