@@ -6,8 +6,7 @@ Copyright   : (c) Eric Mertens 2019
 License     : ISC
 Maintainer  : emertens@gmail.com
 
-These sorting rules are based on the kinds of messages that freenode's
-ircd-seven sends.
+These sorting rules are based on the solanum server notices.
 
 -}
 module Client.Hook.Snotice
@@ -32,12 +31,12 @@ snoticeHook = MessageHook "snotice" True remap
 remap ::
   IrcMsg -> MessageResult
 
-remap (Notice (UserInfo u "" "") _ msg)
+remap (Notice (Source (UserInfo u "" "") _) _ msg)
   | Just msg1 <- Text.stripPrefix "*** Notice -- " msg
   , let msg2 = Text.filter (\x -> x /= '\x02' && x /= '\x0f') msg1
   , Just (lvl, cat) <- characterize msg2
   = if lvl < 1 then OmitMessage
-               else RemapMessage (Notice (UserInfo u "" "*") cat msg1)
+               else RemapMessage (Notice (Source (UserInfo u "" "*") "") cat msg1)
 
 remap _ = PassMessage
 
@@ -73,17 +72,16 @@ patterns = map toPattern
     -- Nick change
     (0, "c", [str|^Nick change: From |]),
     -- Connection limit, more complete regex: ^Too many user connections for [^ ]+![^ ]+@[^ ]+$
-    (0, "u", [str|^Too many user connections for |]),
+    (1, "u", [str|^Too many user connections for |]),
     -- Join alerts, more complete regex: ^User [^ ]+ \([^ ]+@[^ ]+\) trying to join #[^ ]* is a possible spambot$
     (1, "a", [str|^User [^ ]+ \([^ ]+\) trying to join #[^ ]* is a possible spambot|]),
     -- Kline hitting user
     (1, "k", [str|^K/DLINE active for|]),
     -- Connection limit, more complete regex: ^Too many local connections for [^ ]+![^ ]+@[^ ]+$
-    (0, "u", [str|^Too many local connections for |]),
+    (1, "u", [str|^Too many local connections for |]),
     -- Global kline added, more complete regex: ^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} added global [0-9]+ min. K-Line for \[[^ ]+\] \[.*\]$
-    (1, "k", [str|^[^ @]+@freenode/utility-bot/sigyn\{[^ ]+ added global [0-9]+ min. K-Line for |]),
-    (1, "k", [str|^[^ @]+@freenode/bot/proxyscan\{[^ ]+ added global [0-9]+ min. K-Line for |]),
     (2, "k", [str|^[^ ]+ added global [0-9]+ min. K-Line for |]),
+    (2, "k", [str|^[^ ]+ added global [0-9]+ min. X-Line for |]),
     -- Global kline expiring, more complete regex: ^Propagated ban for \[[^ ]+\] expired$
     (0, "k", [str|^Propagated ban for |]),
     -- Chancreate
@@ -102,10 +100,8 @@ patterns = map toPattern
     (1, "m", [str|^Nick collision on|]),
     -- KILLs
     (0, "k", [str|^Received KILL message for [^ ]+\. From NickServ |]),
-    (0, "k", [str|^Received KILL message for [^ ]+\. From [^ .]+\.freenode\.net \(Nickname regained by services\)|]),
     (0, "k", [str|^Received KILL message for [^ ]+\. From syn Path: [^ ]+ \(Facility Blocked\)|]),
     (1, "k", [str|^Received KILL message for [^ ]+\. From syn Path: [^ ]+ \(Banned\)|]),
-    (1, "k", [str|^Received KILL message for [^ ]+\. From Sigyn Path: [^ ]+ \(Spam is off topic on freenode\.\)|]),
     (2, "k", [str|^Received KILL message|]),
     -- Teporary kline expiring, more complete regex: ^Temporary K-line for \[[^ ]+\] expired$
     (0, "k", [str|^Temporary K-line for |]),
@@ -113,16 +109,24 @@ patterns = map toPattern
     -- PATTERN LIST, uncommon snotes. regex performance isn't very important beyond this point
     (2, "a", [str|^Possible Flooder|]),
     (0, "a", [str|^New Max Local Clients: [0-9]+$|]),
+    (1, "a", [str|^Excessive target change from|]),
 
     (1, "f", [str|^Failed (OPER|CHALLENGE) attempt - host mismatch|]),
     (3, "f", [str|^Failed (OPER|CHALLENGE) attempt|]), -- ORDER IMPORTANT - catch all failed attempts that aren't host mismatch
 
+    (1, "k", [str|^Rejecting [XK]-Lined user|]),
+    (1, "k", [str|^Disconnecting [XK]-Lined user|]),
     (1, "k", [str|^KLINE active for|]),
+    (1, "k", [str|^XLINE active for|]),
     (3, "k", [str|^KLINE over-ruled for |]),
     (2, "k", [str|^[^ ]+ added global [0-9]+ min. K-Line from [^ ]+![^ ]+@[^ ]+\{[^ ]+\} for \[[^ ]+\] \[.*\]$|]),
+    (2, "k", [str|^[^ ]+ added global [0-9]+ min. X-Line from [^ ]+![^ ]+@[^ ]+\{[^ ]+\} for \[[^ ]+\] \[.*\]$|]),
     (2, "k", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} has removed the global K-Line for: \[.*\]$|]),
     (2, "k", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} has removed the temporary K-Line for: \[.*\]$|]),
     (2, "k", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} added temporary [0-9]+ min. D-Line for \[[^ ]+\] \[.*\]$|]),
+    (2, "k", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} has removed the X-Line for:|]),
+    (2, "k", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} is removing the X-Line for|]),
+    (2, "k", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} has removed the temporary D-Line for:|]),
 
     (0, "m", [str|^Received SAVE message for|]),
     (0, "m", [str|^Ignored noop SAVE message for|]),
@@ -136,6 +140,7 @@ patterns = map toPattern
     (1, "m", [str|^Rehashing .* by request of system console\.$|]),
     (2, "m", [str|^Updating database by request of [^ ]+( \([^ ]+\))?\.$|]),
     (2, "m", [str|^Rehashing .* by request of [^ ]+( \([^ ]+\))?\.$|]),
+    (2, "m", [str|.* is rehashing server config file$|]),
     (3, "m", [str|^".*", line [0-9+]|]), -- configuration syntax error!
     (0, "m", [str|^Ignoring attempt from [^ ]+( \([^ ]+\))? to set login name for|]),
     (1, "m", [str|^binding listener socket: 99 \(Cannot assign requested address\)$|]),
@@ -143,6 +148,7 @@ patterns = map toPattern
 
     (2, "o", [str|^OPERSPY [^ ]+![^ ]+@[^ ]+\{[^ ]+\} |]),
     (2, "o", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} is overriding |]),
+    (2, "o", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} is using oper-override on |]),
     (2, "o", [str|^[^ ]+ \([^ ]+@[^ ]+\) is now an operator$|]),
     (1, "o", [str|^[^ ]+( \([^ ]+\))? dropped the nick |]),
     (1, "o", [str|^[^ ]+( \([^ ]+\))? dropped the account |]),
@@ -173,6 +179,7 @@ patterns = map toPattern
     (1, "o", [str|^Not kicking immune user |]),
     (1, "o", [str|^Not kicking oper |]),
     (1, "o", [str|^Overriding KICK from |]),
+    (1, "o", [str|^Overriding REMOVE from |]),
     (1, "o", [str|^Server [^ ]+ split from |]),
     (3, "o", [str|^Netsplit [^ ]+ <->|]),
     (2, "o", [str|^Remote SQUIT|]),
@@ -198,12 +205,13 @@ patterns = map toPattern
     (2, "o", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} added X-Line for \[.*\]|]),
     (2, "o", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} is adding a permanent RESV for \[.*\]|]),
     (2, "o", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} added RESV for \[.*\]|]),
+    (2, "o", [str|^[^ ]+![^ ]+@[^ ]+\{[^ ]+\} has removed the RESV for:|]),
     (3, "o", [str|^[^ ]+ is an idiot\. Dropping |]), -- someone k-lined *@*
     (3, "o", [str|^Rejecting email for |]), -- registering from a badmailed address won't trigger this, emailing broken?
     (3, "o", [str|^ERROR |]),
     (3, "o", [str|^No response from [^ ]+, closing link$|]),
 
-    (0, "u", [str|^Too many global connections for [^ ]+![^ ]+@[^ ]+$|]),
+    (1, "u", [str|^Too many global connections for [^ ]+![^ ]+@[^ ]+$|]),
     (0, "u", [str|^Invalid username: |]),
     (0, "u", [str|^HTTP Proxy disconnected: |]),
     (2, "u", [str|^Unauthorised client connection from |]),
