@@ -19,14 +19,16 @@ import Data.Text
 import Data.HashMap.Strict
 import Client.Hook
 
-import Client.Hook.FreRelay
+import Client.Hook.DroneBLRelay
+import Client.Hook.Matterbridge
 import Client.Hook.Snotice
 import Client.Hook.Znc.Buffextras
 
 -- | All the available message hooks.
 messageHooks :: HashMap Text ([Text] -> Maybe MessageHook)
 messageHooks = fromList
-  [ ("snotice"   , \_ -> Just snoticeHook)
-  , ("frerelay"  , freRelayHook)
+  [ ("snotice", \_ -> Just snoticeHook)
+  , ("droneblrelay", droneblRelayHook)
   , ("buffextras", buffextrasHook)
+  , ("matterbridge", matterbridgeHook)
   ]
