@@ -20,7 +20,8 @@ module Client.Image.MircFormatting
   ) where
 
 import           Client.Image.PackedImage as I
-import           Control.Applicative ((<|>))
+import           Client.Image.Palette (Palette, palMonospace)
+import           Control.Applicative ((<|>), empty)
 import           Control.Lens
 import           Data.Attoparsec.Text as Parse
 import           Data.Bits
@@ -30,6 +31,7 @@ import           Data.Text (Text)
 import           Graphics.Vty.Attributes
 import           Data.Vector (Vector)
 import qualified Data.Vector as Vector
+import           Numeric (readHex)
 
 makeLensesFor
   [ ("attrForeColor", "foreColorLens")
@@ -38,15 +40,16 @@ makeLensesFor
   ''Attr
 
 -- | Parse mIRC encoded format characters and hide the control characters.
-parseIrcText :: Text -> Image'
+parseIrcText :: Palette -> Text -> Image'
 parseIrcText = parseIrcText' False
 
 -- | Parse mIRC encoded format characters and render the control characters
 -- explicitly. This view is useful when inputting control characters to make
 -- it clear where they are in the text.
-parseIrcText' :: Bool -> Text -> Image'
-parseIrcText' explicit = either plainText id
-                       . parseOnly (pIrcLine explicit defAttr)
+parseIrcText' :: Bool -> Palette -> Text -> Image'
+parseIrcText' explicit pal
+  = either plainText id
+  . parseOnly (pIrcLine pal explicit False defAttr)
 
 data Segment = TextSegment Text | ControlSegment Char
 
@@ -54,22 +57,36 @@ pSegment :: Parser Segment
 pSegment = TextSegment    <$> takeWhile1 (not . isControl)
        <|> ControlSegment <$> satisfy isControl
 
-pIrcLine :: Bool -> Attr -> Parser Image'
-pIrcLine explicit fmt =
+pIrcLine :: Palette -> Bool -> Bool -> Attr -> Parser Image'
+pIrcLine pal explicit mono fmt =
   do seg <- option Nothing (Just <$> pSegment)
      case seg of
        Nothing -> return mempty
        Just (TextSegment txt) ->
-           do rest <- pIrcLine explicit fmt
+           do rest <- pIrcLine pal explicit mono fmt
               return (text' fmt txt <> rest)
        Just (ControlSegment '\^C') ->
-           do (numberText, colorNumbers) <- match pColorNumbers
-              rest <- pIrcLine explicit (applyColors colorNumbers fmt)
+           do (numberText, colorNumbers) <- match (pColorNumbers pColorNumber)
+              rest <- pIrcLine pal explicit mono (applyColors colorNumbers fmt)
               return $ if explicit
                          then controlImage '\^C'
                               <> text' defAttr numberText
                               <> rest
                           else rest
+       Just (ControlSegment '\^D') ->
+           do (numberText, colorNumbers) <- match (pColorNumbers pColorHex)
+              rest <- pIrcLine pal explicit mono (applyColors colorNumbers fmt)
+              return $ if explicit
+                         then controlImage '\^D'
+                              <> text' defAttr numberText
+                              <> rest
+                          else rest
+       Just (ControlSegment '\^Q')
+         | explicit -> (controlImage '\^Q' <>) <$> rest
+         | otherwise -> rest
+         where
+           rest = pIrcLine pal explicit (not mono)
+                    (if mono then defAttr else view palMonospace pal)
        Just (ControlSegment c)
           -- always render control codes that we don't understand
           | isNothing mbFmt' || explicit ->
@@ -78,22 +95,30 @@ pIrcLine explicit fmt =
           | otherwise -> next
           where
             mbFmt' = applyControlEffect c fmt
-            next   = pIrcLine explicit (fromMaybe fmt mbFmt')
+            next   = pIrcLine pal explicit mono (fromMaybe fmt mbFmt')
 
-pColorNumbers :: Parser (Maybe (MaybeDefault Color, Maybe (MaybeDefault Color)))
-pColorNumbers = option Nothing $
-  do n       <- pNumber
-     Just fc <- pure (mircColor n)
-     bc      <- optional $
-                  do m       <- Parse.char ',' *> pNumber
-                     Just bc <- pure (mircColor m)
-                     pure bc
+pColorNumbers ::
+  Parser (MaybeDefault Color) ->
+  Parser (Maybe (MaybeDefault Color, Maybe (MaybeDefault Color)))
+pColorNumbers color = option Nothing $
+  do fc <- color
+     bc <- optional (Parse.char ',' *> color)
      return (Just (fc,bc))
 
+pColorNumber :: Parser (MaybeDefault Color)
+pColorNumber =
+  do d1 <- digit
+     ds <- option [] (return <$> digit)
+     case mircColor (read (d1:ds)) of
+       Just c -> pure c
+       Nothing -> empty
+
+pColorHex :: Parser (MaybeDefault Color)
+pColorHex = SetTo <$> (rgbColor' <$> p <*> p <*> p)
   where
-    pNumber = do d1 <- digit
-                 ds <- option [] (return <$> digit)
-                 return $! read (d1:ds)
+    p = do x <- satisfy isHexDigit
+           y <- satisfy isHexDigit
+           pure (fst (head (readHex [x,y])))
 
 optional :: Parser a -> Parser (Maybe a)
 optional p = option Nothing (Just <$> p)
@@ -147,6 +172,7 @@ applyControlEffect :: Char -> Attr -> Maybe Attr
 applyControlEffect '\^B' attr = Just $! toggleStyle bold attr
 applyControlEffect '\^V' attr = Just $! toggleStyle reverseVideo attr
 applyControlEffect '\^_' attr = Just $! toggleStyle underline attr
+applyControlEffect '\^^' attr = Just $! toggleStyle strikethrough attr
 applyControlEffect '\^]' attr = Just $! toggleStyle italic attr
 applyControlEffect '\^O' _    = Just defAttr
 applyControlEffect _     _    = Nothing
