@@ -14,9 +14,10 @@ window.
 module Client.Image.StatusLine
   ( statusLineImage
   , minorStatusLineImage
+  , clientTitle
   ) where
 
-import           Client.Image.Message (cleanText)
+import           Client.Image.Message (cleanChar, cleanText)
 import           Client.Image.PackedImage
 import           Client.Image.Palette
 import           Client.State
@@ -31,11 +32,18 @@ import           Data.Maybe
 import           Data.HashMap.Strict (HashMap)
 import           Data.Text (Text)
 import qualified Data.Text as Text
+import qualified Data.Text.Lazy as LText
 import           Graphics.Vty.Attributes
 import qualified Graphics.Vty.Image as Vty
-import           Irc.Identifier (Identifier, idText)
+import           Irc.Identifier (idText)
 import           Numeric
 
+clientTitle :: ClientState -> String
+clientTitle st
+  = map cleanChar
+  $ LText.unpack
+  $ "glirc - " <> imageText (viewFocusLabel st (view clientFocus st))
+
 bar :: Image'
 bar = char (withStyle defAttr bold) '─'
 
@@ -57,6 +65,7 @@ statusLineImage w st =
       , nometaImage (view clientFocus st) st
       , scrollImage st
       , filterImage st
+      , lockImage st
       , latency
       ]
 
@@ -122,6 +131,15 @@ filterImage st
     pal  = clientPalette st
     attr = view palError pal
 
+-- | Indicate when the client editor is locked
+lockImage :: ClientState -> Image'
+lockImage st
+  | view clientEditLock st = infoBubble (string attr "locked")
+  | otherwise              = mempty
+  where
+    pal  = clientPalette st
+    attr = view palError pal
+
 
 -- | Indicate the current connection health. This will either indicate
 -- that the connection is being established or that a ping has been
@@ -147,7 +165,7 @@ latencyImage st = either id id $
 
        PingSent {} -> latencyBubble "wait"
 
-       PingConnecting n _ ->
+       PingConnecting n _ _ ->
          infoBubble (string (view palLatency pal) "connecting" <> retryImage n)
 
        PingNone -> mempty -- just connected no ping sent yet
@@ -199,15 +217,17 @@ activitySummary st
                       Vty.horizCat indicators Vty.<|>
                       Vty.string defAttr "]"
   where
-    winNames = clientWindowNames st ++ repeat '?'
-
-    indicators = foldr aux [] (zip winNames windows)
+    indicators = foldr aux [] windows
     windows    = views clientWindows Map.elems st
 
-    aux (i,w) rest =
+    aux w rest =
+      let name = case view winName w of
+                   Nothing -> '?'
+                   Just i -> i in
+      if view winSilent w then rest else
       case view winMention w of
-        WLImportant -> Vty.char (view palMention  pal) i : rest
-        WLNormal    -> Vty.char (view palActivity pal) i : rest
+        WLImportant -> Vty.char (view palMention  pal) name : rest
+        WLNormal    -> Vty.char (view palActivity pal) name : rest
         WLBoring    -> rest
       where
         pal = clientPalette st
@@ -215,27 +235,28 @@ activitySummary st
 -- | Multi-line activity information enabled by F3
 activityBarImages :: ClientState -> [Vty.Image]
 activityBarImages st
-  = catMaybes
-  $ zipWith baraux winNames
-  $ Map.toList
+  = mapMaybe baraux
+  $ Map.toAscList
   $ view clientWindows st
 
   where
-
-    winNames = clientWindowNames st ++ repeat '?'
-
-    baraux i (focus,w)
+    baraux (focus,w)
+      | view winSilent w = Nothing
       | n == 0 = Nothing -- todo: make configurable
       | otherwise = Just
                   $ unpackImage bar Vty.<|>
                     Vty.char defAttr '[' Vty.<|>
-                    Vty.char (view palWindowName pal) i Vty.<|>
-                    Vty.char defAttr ':' Vty.<|>
-                    Vty.text' (view palLabel pal) focusText Vty.<|>
+                    jumpLabel Vty.<|>
+                    Vty.text' (view palLabel pal) (cleanText focusText) Vty.<|>
                     Vty.char defAttr ':' Vty.<|>
                     Vty.string attr (show n) Vty.<|>
                     Vty.char defAttr ']'
       where
+        jumpLabel =
+          case view winName w of
+            Nothing   -> mempty
+            Just name -> Vty.char (view palWindowName pal) name Vty.<|>
+                         Vty.char defAttr ':'
         n   = view winUnread w
         pal = clientPalette st
         attr = case view winMention w of
@@ -275,32 +296,30 @@ makeLines w (x:xs) = go x xs
 myNickImage :: ClientState -> Vty.Image
 myNickImage st =
   case view clientFocus st of
-    NetworkFocus network      -> nickPart network Nothing
-    ChannelFocus network chan -> nickPart network (Just chan)
+    NetworkFocus network      -> nickPart network
+    ChannelFocus network _    -> nickPart network
     Unfocused                 -> Vty.emptyImage
   where
     pal = clientPalette st
-    nickPart network mbChan =
+    nickPart network =
       case preview (clientConnection network) st of
         Nothing -> Vty.emptyImage
-        Just cs -> Vty.string (view palSigil pal) myChanModes
-           Vty.<|> Vty.text' defAttr (idText nick)
+        Just cs -> Vty.text' attr (cleanText (idText nick))
            Vty.<|> parens defAttr
                      (unpackImage $
                       modesImage (view palUModes pal) (view csModes cs) <>
                       snomaskImage)
           where
+            attr
+              | view csAway cs = view palAway pal
+              | otherwise      = defAttr
+
             nick = view csNick cs
 
             snomaskImage
               | null (view csSnomask cs) = ""
               | otherwise                = " " <> modesImage (view palSnomask pal) (view csSnomask cs)
 
-            myChanModes =
-              case mbChan of
-                Nothing   -> []
-                Just chan -> view (csChannels . ix chan . chanUsers . ix nick) cs
-
 modesImage :: HashMap Char Attr -> String -> Image'
 modesImage pal modes = "+" <> foldMap modeImage modes
   where
@@ -313,19 +332,14 @@ subfocusImage subfocus st = foldMap infoBubble (viewSubfocusLabel pal subfocus)
     pal         = clientPalette st
 
 focusImage :: Focus -> ClientState -> Image'
-focusImage focus st = infoBubble $ mconcat
-    [ char (view palWindowName pal) windowName
-    , char defAttr ':'
-    , viewFocusLabel st focus
-    ]
+focusImage focus st =
+  infoBubble $
+  case preview (clientWindows . ix focus . winName . _Just) st of
+    Nothing -> label
+    Just n  -> char (view palWindowName pal) n <> ":" <> label
   where
     !pal        = clientPalette st
-    windowNames = clientWindowNames st
-
-    windowName = fromMaybe '?'
-               $ do i <- Map.lookupIndex focus (view clientWindows st)
-                    preview (ix i) windowNames
-
+    label       = viewFocusLabel st focus
 
 parens :: Attr -> Vty.Image -> Vty.Image
 parens attr i = Vty.char attr '(' Vty.<|> i Vty.<|> Vty.char attr ')'
@@ -337,20 +351,27 @@ viewFocusLabel st focus =
     Unfocused ->
       char (view palError pal) '*'
     NetworkFocus network ->
-      text' (view palLabel pal) network
+      text' (view palLabel pal) (cleanText network)
     ChannelFocus network channel ->
-      text' (view palLabel pal) network <>
+      text' (view palLabel pal) (cleanText network) <>
       char defAttr ':' <>
-      text' (view palLabel pal) (idText channel) <>
-      channelModesImage network channel st
+      string (view palSigil pal) (cleanChar <$> sigils) <>
+      text' (view palLabel pal) (cleanText (idText channel)) <>
+      channelModes
 
-channelModesImage :: Text -> Identifier -> ClientState -> Image'
-channelModesImage network channel st =
-  case preview (clientConnection network . csChannels . ix channel . chanModes) st of
-    Just modeMap | not (null modeMap) -> " " <> modesImage (view palCModes pal) (Map.keys modeMap)
-    _                                 -> mempty
-  where
-    pal = clientPalette st
+      where
+        (sigils, channelModes) =
+          case preview (clientConnection network) st of
+            Nothing -> ("", mempty)
+            Just cs ->
+               ( let nick = view csNick cs in
+                 view (csChannels . ix channel . chanUsers . ix nick) cs
+
+               , case preview (csChannels . ix channel . chanModes) cs of
+                    Just modeMap | not (null modeMap) ->
+                        " " <> modesImage (view palCModes pal) (Map.keys modeMap)
+                    _ -> mempty
+               )
 
 viewSubfocusLabel :: Palette -> Subfocus -> Maybe Image'
 viewSubfocusLabel pal subfocus =
