@@ -21,32 +21,98 @@ import           Client.Commands.Arguments.Renderer
 import           Client.Commands.Arguments.Parser
 import           Client.Commands.Interpolation
 import           Client.Commands.Recognizer
+import           Client.Image.LineWrap (fullLineWrap, terminate)
 import           Client.Image.Message
 import           Client.Image.MircFormatting
 import           Client.Image.PackedImage
 import           Client.Image.Palette
 import           Client.State
-import           Client.State.Focus
 import qualified Client.State.EditBox as Edit
 import           Control.Lens
-import qualified Data.HashSet as HashSet
-import           Data.HashSet (HashSet)
+import           Data.HashMap.Strict (HashMap)
 import           Data.List
 import qualified Data.Text as Text
 import           Graphics.Vty.Attributes
 import qualified Graphics.Vty.Image as Vty
 import           Irc.Identifier
 
+textboxImage :: Int -> Int -> ClientState -> (Int, Int, Int, Vty.Image) -- ^ cursor column, new offset, image
+textboxImage maxHeight width st =
+  case view clientEditMode st of
+    SingleLineEditor -> singleLineTextboxImage maxHeight width st
+    MultiLineEditor  -> multiLineTextboxImage  maxHeight width st
+
+multiLineTextboxImage :: Int -> Int -> ClientState -> (Int, Int, Int, Vty.Image) -- ^ cursor column, new offset, image
+multiLineTextboxImage _maxHeight width st = (cursorRow, cursorCol, view clientTextBoxOffset st, output)
+  where
+  output = Vty.vertCat (terminate width . unpackImage <$> imgs)
+
+  imgs = as ++ c ++ bs
+
+  content = view (clientTextBox . Edit.content) st
+
+  as  = concatMap (fullLineWrap width)
+      $ mapHead (beginning <>)
+      $ map (<> plainText "\n")
+      $ rndr False <$> reverse (view Edit.above content)
+
+  bs  = concatMap (fullLineWrap width)
+      $ endAfters
+      $ rndr False <$> view Edit.below content
+
+  cur = view Edit.line content
+  curTxt  = view Edit.text cur
+
+  cursorBase
+    = imageWidth $ parseIrcText' True pal
+    $ Text.pack
+    $ take (view Edit.pos cur) curTxt
+
+  (cursorRow, cursorCol) =
+    calcCol (length c + length bs) c (if null as then 1 + cursorBase else cursorBase)
+
+  c = fullLineWrap width
+    $ (if null as then beginning else mempty)
+   <> rndr True curTxt
+   <> (if null bs then ending else plainText "\n")
+
+  rndr = renderLine st pal hilites macros
+  pal = clientPalette st
+  hilites = clientHighlightsFocus (view clientFocus st) st
+  macros = views (clientConfig . configMacros) (fmap macroSpec) st
+
+  attr      = view palTextBox pal
+  beginning = char attr '^'
+  ending    = char attr '$'
+
+  endAfters [] = []
+  endAfters [x] = [x <> ending]
+  endAfters (x:xs) = x <> plainText "\n" : endAfters xs
+
+  -- Using fullLineWrap make calculating the cursor much easier
+  -- to switch to word-breaking lineWrap we'll need some extra
+  -- logic to count skipped spaces.
+  calcCol row [] _ = (row, 0)
+  calcCol row (i:is) n
+    | n < w = (row, n)
+    | otherwise = calcCol (row-1) is (n-w)
+    where
+      w = imageWidth i
+
+mapHead :: (a -> a) -> [a] -> [a]
+mapHead f (x:xs) = f x : xs
+mapHead _ []     = []
+
 -- | Compute the UI image for the text input box. This computes
 -- the logical cursor position on the screen to compensate for
 -- VTY's cursor placement behavior.
-textboxImage :: Int -> ClientState -> (Int, Int, Vty.Image) -- ^ cursor column, new offset, image
-textboxImage width st
-  = (newPos, newOffset, croppedImage)
+singleLineTextboxImage :: Int -> Int -> ClientState -> (Int, Int, Int, Vty.Image) -- ^ cursor column, new offset, image
+singleLineTextboxImage _maxHeight width st
+  = (1, newPos, newOffset, croppedImage)
   where
   macros = views (clientConfig . configMacros) (fmap macroSpec) st
   (txt, content) =
-     renderContent st myNick nicks macros pal
+     renderContent st hilites macros pal
        (view (clientTextBox . Edit.content) st)
 
   lineImage = unpackImage (beginning <> content <> ending)
@@ -76,14 +142,7 @@ textboxImage width st
   beginning = char attr '^'
   ending    = char attr '$'
 
-
-  (myNick,nicks) =
-    case view clientFocus st of
-      ChannelFocus network channel ->
-         (clientHighlightsNetwork network st,
-          HashSet.fromList (channelUserList network channel st)
-         )
-      _ -> (HashSet.empty, HashSet.empty)
+  hilites = clientHighlightsFocus (view clientFocus st) st
 
 
 -- | Renders the whole, uncropped text box as well as the 'String'
@@ -91,13 +150,12 @@ textboxImage width st
 -- the logical cursor position of the cropped version of the text box.
 renderContent ::
   ClientState          {- ^ client state                          -} ->
-  HashSet Identifier   {- ^ my nicknames                          -} ->
-  HashSet Identifier   {- ^ other nicknames                       -} ->
+  HashMap Identifier Highlight {- ^ highlights                    -} ->
   Recognizer MacroSpec {- ^ macro completions                     -} ->
   Palette              {- ^ palette                               -} ->
   Edit.Content         {- ^ content                               -} ->
   (Int, Image')        {- ^ left-of-cursor width, image rendering -}
-renderContent st myNick nicks macros pal c = (leftLen, wholeImg)
+renderContent st hilites macros pal c = (leftLen, wholeImg)
   where
   as  = reverse (view Edit.above c)
   bs  = view Edit.below c
@@ -109,9 +167,9 @@ renderContent st myNick nicks macros pal c = (leftLen, wholeImg)
   -- ["one","two"] "three" --> "two one three"
   leftLen = length leftImgs -- separators
           + sum (map imageWidth leftImgs)
-          + imageWidth (parseIrcText' True (Text.pack leftCur))
+          + imageWidth (parseIrcText' True pal (Text.pack leftCur))
 
-  rndr = renderLine st pal myNick nicks macros
+  rndr = renderLine st pal hilites macros
 
   leftImgs = map (rndr False) as
 
@@ -127,13 +185,12 @@ renderContent st myNick nicks macros pal c = (leftLen, wholeImg)
 renderLine ::
   ClientState ->
   Palette ->
-  HashSet Identifier ->
-  HashSet Identifier ->
+  HashMap Identifier Highlight ->
   Recognizer MacroSpec {- ^ commands     -} ->
   Bool                 {- ^ focused      -} ->
   String               {- ^ input text   -} ->
   Image'               {- ^ output image -}
-renderLine st pal myNick nicks macros focused input =
+renderLine st pal hilites macros focused input =
 
   case span (' '==) input of
     (spcs, '/':xs) -> string defAttr spcs <> char defAttr '/'
@@ -149,7 +206,7 @@ renderLine st pal myNick nicks macros focused input =
 
         allCommands = (Left <$> macros) <> (Right <$> commands)
         (attr, continue)
-          = case recognize (Text.pack cmd) allCommands of
+          = case recognize (Text.toLower (Text.pack cmd)) allCommands of
               Exact (Right Command{cmdArgumentSpec = spec}) ->
                 ( specAttr spec
                 , render pal st focused spec
@@ -160,10 +217,10 @@ renderLine st pal myNick nicks macros focused input =
                 )
               Prefix _ ->
                 ( view palCommandPrefix pal
-                , parseIrcTextWithNicks pal myNick nicks focused . Text.pack
+                , parseIrcTextWithNicks pal hilites focused . Text.pack
                 )
               Invalid ->
                 ( view palCommandError pal
-                , parseIrcTextWithNicks pal myNick nicks focused . Text.pack
+                , parseIrcTextWithNicks pal hilites focused . Text.pack
                 )
-    _ -> parseIrcTextWithNicks pal myNick nicks focused (Text.pack input)
+    _ -> parseIrcTextWithNicks pal hilites focused (Text.pack input)
