@@ -27,6 +27,10 @@ module Client.Message
   -- * Client message operations
   , IrcSummary(..)
   , msgSummary
+  , summaryActor
+
+  -- * Quit message details
+  , QuitKind(..)
 
   -- * Client message operations
   , msgText
@@ -39,7 +43,6 @@ import           Irc.Message
 import           Irc.Identifier
 import           Irc.UserInfo
 import           Irc.Codes
-import           Client.State.DCC (isSend)
 
 data MessageBody
   = IrcBody    !IrcMsg
@@ -56,15 +59,19 @@ data ClientMessage = ClientMessage
 
 makeLenses ''ClientMessage
 
+data QuitKind
+  = NormalQuit -- ^ User quit
+  | MassQuit   -- ^ Mass event like a netsplit
+  deriving (Eq, Show)
+
 data IrcSummary
   = JoinSummary {-# UNPACK #-} !Identifier
-  | QuitSummary {-# UNPACK #-} !Identifier
+  | QuitSummary {-# UNPACK #-} !Identifier !QuitKind
   | PartSummary {-# UNPACK #-} !Identifier
   | NickSummary {-# UNPACK #-} !Identifier {-# UNPACK #-} !Identifier
   | ReplySummary {-# UNPACK #-} !ReplyCode
   | ChatSummary {-# UNPACK #-} !UserInfo
   | CtcpSummary {-# UNPACK #-} !Identifier
-  | DccSendSummary {-# UNPACK #-} !Identifier
   | ChngSummary {-# UNPACK #-} !Identifier -- ^ Chghost command
   | AcctSummary {-# UNPACK #-} !Identifier -- ^ Account command
   | NoSummary
@@ -87,17 +94,37 @@ msgSummary (NormalBody _  ) = NoSummary
 ircSummary :: IrcMsg -> IrcSummary
 ircSummary msg =
   case msg of
-    Join who _ _ _  -> JoinSummary (userNick who)
-    Part who _ _    -> PartSummary (userNick who)
-    Quit who _      -> QuitSummary (userNick who)
-    Nick who who'   -> NickSummary (userNick who) who'
-    Privmsg who _ _ -> ChatSummary who
-    Notice who _ _  -> ChatSummary who
-    Ctcp who _ "ACTION" _ -> ChatSummary who
-    Ctcp who _ "DCC" txt | isSend txt -> DccSendSummary (userNick who)
-    Ctcp who _ _ _ -> CtcpSummary (userNick who)
-    CtcpNotice who _ _ _ -> ChatSummary who
-    Reply code _    -> ReplySummary code
-    Account who _   -> AcctSummary (userNick who)
-    Chghost who _ _ -> ChngSummary (userNick who)
+    Join who _ _ _  -> JoinSummary (userNick (srcUser who))
+    Part who _ _    -> PartSummary (userNick (srcUser who))
+    Quit who mbTxt  -> QuitSummary (userNick (srcUser who)) (quitKind mbTxt)
+    Nick who who'   -> NickSummary (userNick (srcUser who)) who'
+    Privmsg who _ _ -> ChatSummary (srcUser who)
+    Notice who _ _  -> ChatSummary (srcUser who)
+    Ctcp who _ "ACTION" _ -> ChatSummary (srcUser who)
+    Ctcp who _ _ _ -> CtcpSummary (userNick (srcUser who))
+    CtcpNotice who _ _ _ -> ChatSummary (srcUser who)
+    Reply _ code _  -> ReplySummary code
+    Account who _   -> AcctSummary (userNick (srcUser who))
+    Chghost who _ _ -> ChngSummary (userNick (srcUser who))
     _               -> NoSummary
+
+quitKind :: Maybe Text -> QuitKind
+quitKind mbReason =
+  case mbReason of
+    Just "*.net *.split"        -> MassQuit
+    _                           -> NormalQuit
+
+summaryActor :: IrcSummary -> Maybe Identifier
+summaryActor s =
+  case s of
+    JoinSummary who   -> Just who
+    QuitSummary who _ -> Just who
+    PartSummary who   -> Just who
+    NickSummary who _ -> Just who
+    ChatSummary who   -> Just (userNick who)
+    CtcpSummary who   -> Just who
+    AcctSummary who   -> Just who
+    ChngSummary who   -> Just who
+    ReplySummary {}   -> Nothing
+    NoSummary         -> Nothing
+
