@@ -28,6 +28,7 @@ module Client.Network.Async
   , createConnection
   , Client.Network.Async.send
   , Client.Network.Async.recv
+  , upgrade
 
   -- * Abort connections
   , abortConnection
@@ -45,15 +46,19 @@ import           Control.Monad
 import           Data.ByteString (ByteString)
 import qualified Data.ByteString as B
 import           Data.Foldable
-import           Data.Time
 import           Data.List
-import           Irc.RateLimit
-import           Hookup
+import           Data.List.Split (chunksOf)
 import           Data.Text (Text)
 import qualified Data.Text as Text
+import           Data.Time
+import           Data.Traversable
 import           Data.Word (Word8)
+import           Hookup
+import           Hookup.OpenSSL (getPubKeyDer)
+import           Irc.RateLimit
 import           Numeric (showHex)
-import           OpenSSL.X509 (X509, printX509)
+import qualified OpenSSL.EVP.Digest as Digest
+import           OpenSSL.X509 (X509, printX509, writeDerX509)
 
 
 -- | Handle for a network connection
@@ -61,14 +66,22 @@ data NetworkConnection = NetworkConnection
   { connOutQueue :: TQueue ByteString
   , connInQueue  :: TQueue NetworkEvent
   , connAsync    :: Async ()
+  , connUpgrade  :: MVar (IO ())
   }
 
+-- | Signals that the server is ready to initiate the TLS handshake.
+-- This is a no-op when not in a starttls state.
+upgrade :: NetworkConnection -> IO ()
+upgrade c = join (swapMVar (connUpgrade c) (pure ()))
+
 -- | The sum of incoming events from a network connection. All events
 -- are annotated with a network ID matching that given when the connection
 -- was created as well as the time at which the message was recieved.
 data NetworkEvent
   -- | Event for successful connection to host (certificate lines)
-  = NetworkOpen  !ZonedTime [Text]
+  = NetworkOpen  !ZonedTime
+  -- | Event indicating TLS is in effect
+  | NetworkTLS  [Text]
   -- | Event for a new recieved line (newline removed)
   | NetworkLine  !ZonedTime !ByteString
   -- | Report an error on network connection network connection failed
@@ -85,6 +98,7 @@ data TerminationReason
   = PingTimeout      -- ^ sent when ping timer expires
   | ForcedDisconnect -- ^ sent when client commands force disconnect
   | StsUpgrade       -- ^ sent when the client disconnects due to sts policy
+  | StartTLSFailed   -- ^ STARTTLS was expected by server had an error
   | BadCertFingerprint ByteString (Maybe ByteString)
   | BadPubkeyFingerprint ByteString (Maybe ByteString)
   deriving Show
@@ -93,6 +107,7 @@ instance Exception TerminationReason where
   displayException PingTimeout      = "connection killed due to ping timeout"
   displayException ForcedDisconnect = "connection killed by client command"
   displayException StsUpgrade       = "connection killed by sts policy"
+  displayException StartTLSFailed   = "connection killed due to failed STARTTLS"
   displayException (BadCertFingerprint expect got) =
        "Expected certificate fingerprint: " ++ formatDigest expect ++
        "; got: "    ++ maybe "none" formatDigest got
@@ -125,9 +140,12 @@ createConnection delay settings =
    do outQueue <- newTQueueIO
       inQueue  <- newTQueueIO
 
+      upgradeMVar <- newEmptyMVar
+
       supervisor <- async $
                       threadDelay (delay * 1000000) >>
-                      startConnection settings inQueue outQueue
+                      withConnection settings
+                        (startConnection settings inQueue outQueue upgradeMVar)
 
       let recordFailure :: SomeException -> IO ()
           recordFailure ex =
@@ -151,6 +169,7 @@ createConnection delay settings =
         { connOutQueue = outQueue
         , connInQueue  = inQueue
         , connAsync    = supervisor
+        , connUpgrade  = upgradeMVar
         }
 
 
@@ -158,26 +177,56 @@ startConnection ::
   ServerSettings ->
   TQueue NetworkEvent ->
   TQueue ByteString ->
+  MVar (IO ()) ->
+  Connection ->
   IO ()
-startConnection settings inQueue outQueue =
-  do rate <- newRateLimit
-               (view ssFloodPenalty settings)
-               (view ssFloodThreshold settings)
-     withConnection settings $ \h ->
-       do for_ (view ssTlsCertFingerprint settings)
-            (checkCertFingerprint h)
-          for_ (view ssTlsPubkeyFingerprint settings)
-            (checkPubkeyFingerprint h)
+startConnection settings inQueue outQueue upgradeMVar h =
+  do reportNetworkOpen inQueue
+     ready <- presend
+     when ready $
+       do checkFingerprints
+          race_ receiveMain sendMain
+  where
+    receiveMain = receiveLoop h inQueue
 
-          reportNetworkOpen h inQueue
-          withAsync (sendLoop h outQueue rate) $ \sender ->
-            withAsync (receiveLoop h inQueue) $ \receiver ->
-              do res <- waitEitherCatch sender receiver
-                 case res of
-                   Left  Right{}  -> fail "PANIC: sendLoop returned"
-                   Right Right{}  -> return ()
-                   Left  (Left e) -> throwIO e
-                   Right (Left e) -> throwIO e
+    sendMain =
+      do rate <- newRateLimit (view ssFloodPenalty   settings)
+                              (view ssFloodThreshold settings)
+         sendLoop h outQueue rate
+
+    presend =
+      case view ssTls settings of
+        TlsNo  -> True <$ putMVar upgradeMVar (pure ())
+        TlsYes ->
+          do txts <- describeCertificates h
+             putMVar upgradeMVar (pure ())
+             atomically (writeTQueue inQueue (NetworkTLS txts))
+             pure True
+        TlsStart ->
+          do Hookup.send h "STARTTLS\n"
+             r <- withAsync receiveMain $ \t ->
+                    do putMVar upgradeMVar (cancel t)
+                       waitCatch t
+             case r of
+               -- network connection closed
+               Right () -> pure False
+
+               -- pre-receiver was killed by a call to 'upgrade'
+               Left e | Just AsyncCancelled <- fromException e ->
+                  do Hookup.upgradeTls (tlsParams settings) (view ssHostName settings) h
+                     txts <- describeCertificates h
+                     atomically (writeTQueue inQueue (NetworkTLS txts))
+                     pure True
+
+               -- something else went wrong with network IO
+               Left e -> throwIO e
+
+    checkFingerprints =
+      case view ssTls settings of
+        TlsNo -> pure ()
+        _ ->
+          do for_ (view ssTlsCertFingerprint   settings) (checkCertFingerprint   h)
+             for_ (view ssTlsPubkeyFingerprint settings) (checkPubkeyFingerprint h)
 
 checkCertFingerprint :: Connection -> Fingerprint -> IO ()
 checkCertFingerprint h fp =
@@ -199,15 +248,18 @@ checkPubkeyFingerprint h fp =
      unless (Just expect == got)
        (throwIO (BadPubkeyFingerprint expect got))
 
-reportNetworkOpen :: Connection -> TQueue NetworkEvent -> IO ()
-reportNetworkOpen h inQueue =
+reportNetworkOpen :: TQueue NetworkEvent -> IO ()
+reportNetworkOpen inQueue =
   do now <- getZonedTime
-     mbServer <- getPeerCertificate h
-     let mbClient = getClientCertificate h
+     atomically (writeTQueue inQueue (NetworkOpen now))
+
+describeCertificates :: Connection -> IO [Text]
+describeCertificates h =
+  do mbServer <- getPeerCertificate h
+     mbClient <- getClientCertificate h
      cTxts <- certText "Server" mbServer
      sTxts <- certText "Client" mbClient
-     let txts = cTxts ++ sTxts
-     atomically (writeTQueue inQueue (NetworkOpen now txts))
+     pure (reverse (cTxts ++ sTxts))
 
 certText :: String -> Maybe X509 -> IO [Text]
 certText label mbX509 =
@@ -215,7 +267,34 @@ certText label mbX509 =
     Nothing -> pure []
     Just x509 ->
       do str <- printX509 x509
-         return $! reverse (Text.lines (Text.pack ("<<" ++ label ++ ">>\n" ++ str)))
+         fps <- getFingerprints x509
+         pure $ map Text.pack
+              $ ('\^B' : label)
+              : map colorize (lines str ++ fps)
+  where
+    colorize x@(' ':_) = x
+    colorize xs = "\^C07" ++ xs
+
+getFingerprints :: X509 -> IO [String]
+getFingerprints x509 =
+  do certDer <- writeDerX509 x509
+     spkiDer <- getPubKeyDer x509
+     xss <- for ["sha1", "sha256", "sha512"] $ \alg ->
+              do mb <- Digest.getDigestByName alg
+                 pure $ case mb of
+                   Nothing -> []
+                   Just d ->
+                       ("Certificate " ++ alg ++ " fingerprint:")
+                     : fingerprintLines (Digest.digestLBS d certDer)
+                    ++ ("SPKI " ++ alg ++ " fingerprint:")
+                     : fingerprintLines (Digest.digestBS d spkiDer)
+     pure (concat xss)
+
+fingerprintLines :: ByteString -> [String]
+fingerprintLines
+  = map ("    "++)
+  . chunksOf (16*3)
+  . formatDigest
 
 formatDigest :: ByteString -> String
 formatDigest
@@ -228,12 +307,12 @@ showByte x
   | x < 0x10  = '0' : showHex x ""
   | otherwise = showHex x ""
 
-sendLoop :: Connection -> TQueue ByteString -> RateLimit -> IO ()
+sendLoop :: Connection -> TQueue ByteString -> RateLimit -> IO a
 sendLoop h outQueue rate =
   forever $
-    do msg <- atomically (readTQueue outQueue)
-       tickRateLimit rate
-       Hookup.send h msg
+  do msg <- atomically (readTQueue outQueue)
+     tickRateLimit rate
+     Hookup.send h msg
 
 ircMaxMessageLength :: Int
 ircMaxMessageLength = 512
