@@ -20,8 +20,7 @@ module Client.State
   , clientTextBox
   , clientTextBoxOffset
   , clientConnections
-  , clientDCC
-  , clientDCCUpdates
+  , clientThreadJoins
   , clientWidth
   , clientHeight
   , clientEvents
@@ -44,9 +43,12 @@ module Client.State
   , clientActivityReturn
   , clientErrorMsg
   , clientLayout
+  , clientEditMode
+  , clientEditLock
   , clientRtsStats
   , clientConfigPath
   , clientStsPolicy
+  , clientHighlights
 
   -- * Client operations
   , withClientState
@@ -54,7 +56,6 @@ module Client.State
   , clientFilter
   , buildMatcher
   , clientToggleHideMeta
-  , clientHighlightsNetwork
   , channelUserList
 
   , consumeInput
@@ -66,13 +67,13 @@ module Client.State
   , addConnection
   , removeNetwork
   , clientTick
-  , queueDCCTransfer
   , applyMessageToClientState
-  , clientHighlights
+  , clientHighlightsFocus
   , clientWindowNames
   , clientPalette
   , clientAutoconnects
   , clientActiveCommand
+  , clientNextWindowName
 
   , clientExtraFocuses
   , currentNickCompletionMode
@@ -82,6 +83,7 @@ module Client.State
   , recordNetworkMessage
   , recordError
   , recordIrcMessage
+  , recordSuccess
 
   -- * Focus manipulation
   , changeFocus
@@ -125,7 +127,6 @@ import qualified Client.State.EditBox as Edit
 import           Client.State.Focus
 import           Client.State.Network
 import           Client.State.Window
-import           Client.State.DCC
 import           ContextFilter
 import           Control.Applicative
 import           Control.Concurrent.MVar
@@ -156,7 +157,7 @@ import           Irc.RawIrcMsg
 import           Irc.UserInfo
 import           LensUtils
 import           RtsStats (Stats)
-import qualified System.Random.MWC as Random
+import qualified System.Random as Random
 import           Text.Regex.TDFA
 import           Text.Regex.TDFA.String (compile)
 
@@ -172,8 +173,7 @@ data ClientState = ClientState
 
   , _clientConnections       :: !(HashMap Text NetworkState) -- ^ state of active connections
   , _clientEvents            :: !(TQueue NetworkEvent)    -- ^ incoming network event queue
-  , _clientDCC               :: !DCCState                 -- ^ DCC subsystem
-  , _clientDCCUpdates        :: !(TChan DCCUpdate)        -- ^ DCC update events
+  , _clientThreadJoins       :: TQueue (Int, ThreadEntry) -- ^ Finished threads ready to report
 
   , _clientConfig            :: !Configuration            -- ^ client configuration
   , _clientConfigPath        :: !FilePath                 -- ^ client configuration file path
@@ -188,7 +188,9 @@ data ClientState = ClientState
   , _clientActivityBar       :: !Bool                     -- ^ visible activity bar
   , _clientShowPing          :: !Bool                     -- ^ visible ping time
   , _clientRegex             :: Maybe Matcher             -- ^ optional persistent filter
-  , _clientLayout            :: !LayoutMode               -- ^ layout mode for split screen
+  , _clientLayout            :: LayoutMode                -- ^ layout mode for split screen
+  , _clientEditMode          :: EditMode                  -- ^ editor rendering mode
+  , _clientEditLock          :: Bool                      -- ^ editor locked and won't send
 
   , _clientBell              :: !Bool                     -- ^ sound a bell next draw
 
@@ -201,11 +203,13 @@ data ClientState = ClientState
   , _clientRtsStats          :: Maybe Stats               -- ^ most recent GHC RTS stats
 
   , _clientStsPolicy         :: !(HashMap Text StsPolicy) -- ^ STS policy entries
+  , _clientHighlights        :: !(HashMap Identifier Highlight) -- ^ highlights
   }
 
 data Matcher = Matcher
   { matcherBefore :: !Int
   , matcherAfter  :: !Int
+  , matcherMax    :: Maybe Int
   , matcherPred   :: LText.Text -> Bool
   }
 
@@ -249,7 +253,7 @@ withClientState cfgPath cfg k =
   withExtensionState $ \exts ->
 
   do events    <- atomically newTQueue
-     dccEvents <- atomically newTChan
+     threadQueue <- atomically newTQueue
      sts       <- readPolicyFile
      let ignoreIds = map mkId (view configIgnores cfg)
      k ClientState
@@ -257,8 +261,7 @@ withClientState cfgPath cfg k =
         , _clientIgnores           = HashSet.fromList ignoreIds
         , _clientIgnoreMask        = buildMask ignoreIds
         , _clientConnections       = _Empty # ()
-        , _clientDCC               = emptyDCCState
-        , _clientDCCUpdates        = dccEvents
+        , _clientThreadJoins       = threadQueue
         , _clientTextBox           = Edit.defaultEditBox
         , _clientTextBoxOffset     = 0
         , _clientWidth             = 80
@@ -275,6 +278,8 @@ withClientState cfgPath cfg k =
         , _clientDetailView        = False
         , _clientRegex             = Nothing
         , _clientLayout            = view configLayout cfg
+        , _clientEditMode          = SingleLineEditor
+        , _clientEditLock          = False
         , _clientActivityBar       = view configActivityBar cfg
         , _clientShowPing          = view configShowPing cfg
         , _clientBell              = False
@@ -283,6 +288,7 @@ withClientState cfgPath cfg k =
         , _clientErrorMsg          = Nothing
         , _clientRtsStats          = Nothing
         , _clientStsPolicy         = sts
+        , _clientHighlights        = HashMap.empty
         }
 
 withExtensionState :: (ExtensionState -> IO a) -> IO a
@@ -308,6 +314,13 @@ abortNetwork network st =
                   -- unassociate this network name from this network id
                   return $! over clientConnections (sans network) st
 
+recordSuccess :: ZonedTime -> ClientState -> Text -> ClientState
+recordSuccess now ste m =
+  recordNetworkMessage ClientMessage
+    { _msgTime    = now
+    , _msgBody    = NormalBody m
+    , _msgNetwork = ""
+    } ste
 
 -- | Add a message to the window associated with a given channel
 recordChannelMessage ::
@@ -317,7 +330,7 @@ recordChannelMessage ::
   ClientState ->
   ClientState
 recordChannelMessage network channel msg st
-  = recordLogLine msg channel
+  = recordLogLine msg statusModes channel'
   $ recordWindowLine focus wl st
   where
     focus      = ChannelFocus network channel'
@@ -326,8 +339,7 @@ recordChannelMessage network channel msg st
     rendParams = MessageRendererParams
       { rendStatusMsg   = statusModes
       , rendUserSigils  = computeMsgLineSigils network channel' msg st
-      , rendNicks       = HashSet.fromList (channelUserList network channel' st)
-      , rendMyNicks     = highlights
+      , rendHighlights  = highlights
       , rendPalette     = clientPalette st
       , rendAccounts    = accounts
       }
@@ -337,7 +349,7 @@ recordChannelMessage network channel msg st
     possibleStatusModes     = view csStatusMsg cs
     (statusModes, channel') = splitStatusMsgModes possibleStatusModes channel
     importance              = msgImportance msg st
-    highlights              = clientHighlightsNetwork network st
+    highlights              = clientHighlightsFocus (ChannelFocus network channel) st
 
     accounts =
       if view (csSettings . ssShowAccounts) cs
@@ -347,14 +359,15 @@ recordChannelMessage network channel msg st
 
 recordLogLine ::
   ClientMessage {- ^ message      -} ->
+  [Char]        {- ^ status modes -} ->
   Identifier    {- ^ target       -} ->
   ClientState   {- ^ client state -} ->
   ClientState
-recordLogLine msg target st =
+recordLogLine msg statusModes target st =
   case view (clientConnection (view msgNetwork msg) . csSettings . ssLogDir) st of
     Nothing -> st
     Just dir ->
-      case renderLogLine msg dir target of
+      case renderLogLine msg dir statusModes target of
         Nothing  -> st
         Just ll  -> over clientLogQueue (cons ll) st
 
@@ -375,10 +388,10 @@ msgImportance :: ClientMessage -> ClientState -> WindowLineImportance
 msgImportance msg st =
   let network = view msgNetwork msg
       me      = preview (clientConnection network . csNick) st
-      highlights = clientHighlightsNetwork network st
+      highlights = clientHighlightsFocus (NetworkFocus network) st
       isMe x  = Just x == me
       checkTxt txt
-        | any (\x -> HashSet.member (mkId x) highlights)
+        | any (\x -> Just HighlightMe == HashMap.lookup (mkId x) highlights)
               (nickSplit txt) = WLImportant
         | otherwise           = WLNormal
   in
@@ -400,18 +413,19 @@ msgImportance msg st =
           | isMe tgt  -> WLImportant
           | otherwise -> checkTxt txt
         Ctcp{} -> WLNormal
-        Part who _ _ | isMe (userNick who) -> WLImportant
-                     | otherwise           -> WLBoring
+        Wallops{} -> WLImportant
+        Part who _ _ | isMe (userNick (srcUser who)) -> WLImportant
+                     | otherwise -> WLBoring
         Kick _ _ kicked _ | isMe kicked -> WLImportant
                           | otherwise   -> WLNormal
         Error{} -> WLImportant
 
         -- channel information
-        Reply RPL_TOPIC _        -> WLBoring
-        Reply RPL_INVITING _     -> WLBoring
+        Reply _ RPL_TOPIC _    -> WLBoring
+        Reply _ RPL_INVITING _ -> WLBoring
 
         -- remaining replies go to network window
-        Reply cmd _ ->
+        Reply _ cmd _ ->
           case replyCodeType (replyCodeInfo cmd) of
             ErrorReply -> WLImportant
             _          -> WLNormal
@@ -432,8 +446,8 @@ ircIgnorable msg !st =
     _                    -> Nothing
   where
     checkUser !who
-      | identIgnored who st = Just (userNick who)
-      | otherwise           = Nothing
+      | identIgnored (srcUser who) st = Just (userNick (srcUser who))
+      | otherwise = Nothing
 
 
 
@@ -454,7 +468,6 @@ recordIrcMessage ::
 recordIrcMessage network target msg st =
   updateTransientError (NetworkFocus network) msg $
   case target of
-    TargetHidden      -> st
     TargetNetwork     -> recordNetworkMessage msg st
     TargetWindow chan -> recordChannelMessage network chan msg st
     TargetUser user   ->
@@ -463,8 +476,7 @@ recordIrcMessage network target msg st =
                              (addToWindow wl) st')
            st chans
       where
-        cfg   = view clientConfig st
-        wl    = toWindowLine' cfg WLBoring msg
+        wl    = toWindowLine' network st WLBoring msg
         chans = user
               : case preview (clientConnection network . csChannels) st of
                   Nothing -> []
@@ -480,7 +492,7 @@ computeMsgLineSigils ::
   [Char] {- ^ sigils -}
 computeMsgLineSigils network channel msg st =
   case msgActor =<< preview (msgBody . _IrcBody) msg of
-    Just user -> computeUserSigils network channel (userNick user) st
+    Just user -> computeUserSigils network channel (userNick (srcUser user)) st
     Nothing   -> []
 
 -- | Compute sigils for a user on a channel
@@ -510,7 +522,7 @@ updateTransientError destination msg st
   case view msgBody msg of
     ErrorBody txt       -> err txt
     IrcBody (Error txt) -> err txt
-    IrcBody (Reply code args)
+    IrcBody (Reply _ code args)
       | let info = replyCodeInfo code
       , ErrorReply <- replyCodeType info ->
           err (Text.intercalate " " (replyCodeText info : drop 1 args))
@@ -526,9 +538,7 @@ recordNetworkMessage msg st = updateTransientError focus msg
     focus      | Text.null network = Unfocused
                | otherwise         = NetworkFocus (view msgNetwork msg)
     importance = msgImportance msg st
-    wl         = toWindowLine' cfg importance msg
-
-    cfg        = view clientConfig st
+    wl         = toWindowLine' network st importance msg
 
 recordError ::
   ZonedTime       {- ^ now             -} ->
@@ -543,6 +553,14 @@ recordError now net msg =
     , _msgBody    = ErrorBody msg
     }
 
+clientNextWindowName :: ClientState -> Char
+clientNextWindowName st =
+  case clientWindowNames st \\ usedNames of
+    []  -> '\0'
+    c:_ -> c
+  where
+    usedNames = toListOf (clientWindows . folded . winName . _Just) st
+
 -- | Record window line at the given focus creating the window if necessary
 recordWindowLine ::
   Focus ->
@@ -551,7 +569,11 @@ recordWindowLine ::
   ClientState
 recordWindowLine focus wl st = st2
   where
-    freshWindow = emptyWindow { _winHideMeta = view (clientConfig . configHideMeta) st }
+    freshWindow = emptyWindow
+      { _winName' = clientNextWindowName st
+      , _winHideMeta = view (clientConfig . configHideMeta) st
+      }
+
     st1 = over (clientWindows . at focus)
                (\w -> Just $! addToWindow wl (fromMaybe freshWindow w))
                st
@@ -579,11 +601,11 @@ toWindowLine params importance msg = WindowLine
     (prefix, image, full) = msgImage (view msgTime msg) params (view msgBody msg)
 
 -- | 'toWindowLine' but with mostly defaulted parameters.
-toWindowLine' :: Configuration -> WindowLineImportance -> ClientMessage -> WindowLine
-toWindowLine' config =
+toWindowLine' :: Text -> ClientState -> WindowLineImportance -> ClientMessage -> WindowLine
+toWindowLine' network st =
   toWindowLine defaultRenderParams
-    { rendPalette     = view configPalette     config
-    , rendMyNicks     = view configExtraHighlights config
+    { rendPalette     = view (clientConfig . configPalette) st
+    , rendHighlights  = clientHighlightsFocus (NetworkFocus network) st
     }
 
 
@@ -660,18 +682,23 @@ clientFilter :: ClientState -> (a -> LText.Text) -> [a] -> [a]
 clientFilter st f xs =
   case clientMatcher st of
     Nothing -> xs
-    Just m  ->
+    Just m ->
+      limit $
       filterContext
+        (matcherAfter m) -- client messages are stored in descending order
         (matcherBefore m)
-        (matcherAfter m)
         (matcherPred m . f)
         xs
+     where
+       limit = maybe id take (matcherMax m)
 
 data MatcherArgs = MatcherArgs
   { argAfter     :: !Int
   , argBefore    :: !Int
   , argInvert    :: !Bool
   , argSensitive :: !Bool
+  , argMax       :: Maybe Int
+  , argPlain     :: !Bool
   }
 
 defaultMatcherArgs :: MatcherArgs
@@ -680,6 +707,8 @@ defaultMatcherArgs = MatcherArgs
   , argBefore    = 0
   , argInvert    = False
   , argSensitive = True
+  , argMax       = Nothing
+  , argPlain     = False
   }
 
 buildMatcher :: String -> Maybe Matcher
@@ -689,20 +718,29 @@ buildMatcher = go defaultMatcherArgs
       case dropWhile (' '==) reStr of
         '-' : 'i' : ' ' : reStr' -> go args{argSensitive=False} reStr'
         '-' : 'v' : ' ' : reStr' -> go args{argInvert=True} reStr'
+        '-' : 'F' : ' ' : reStr' -> go args{argPlain=True} reStr'
         '-' : 'A' : reStr' | [(a,' ':reStr'')] <- reads reStr', a>=0 -> go args{argAfter=a} reStr''
         '-' : 'B' : reStr' | [(b,' ':reStr'')] <- reads reStr', b>=0 -> go args{argBefore=b} reStr''
         '-' : 'C' : reStr' | [(c,' ':reStr'')] <- reads reStr', c>=0 -> go args{argAfter=c,argBefore=c} reStr''
+        '-' : 'm' : reStr' | [(m,' ':reStr'')] <- reads reStr', m>=0 -> go args{argMax=Just m} reStr''
         '-' : '-' : ' ' : reStr' -> finish args reStr'
         _ -> finish args reStr
 
-    finish args reStr =
-      case compile defaultCompOpt{caseSensitive=argSensitive args}
-                   defaultExecOpt{captureGroups=False}
-                   reStr of
-        Left{}  -> Nothing
-        Right r
-          | argInvert args -> Just (Matcher (argBefore args) (argAfter args) (not . matchTest r . LText.unpack))
-          | otherwise      -> Just (Matcher (argBefore args) (argAfter args) (      matchTest r . LText.unpack))
+    finish args reStr
+      | argPlain args =
+          if argSensitive args
+          then matcher (LText.isInfixOf (LText.fromStrict (Text.pack reStr)))
+          else matcher (LText.isInfixOf (LText.fromStrict (Text.toLower (Text.pack reStr))) . LText.toLower)
+      | otherwise =
+        case compile defaultCompOpt{caseSensitive=argSensitive args}
+                     defaultExecOpt{captureGroups=False}
+                     reStr of
+          Left{}  -> Nothing
+          Right r -> matcher (matchTest r . LText.unpack)
+      where
+        matcher f
+          | argInvert args = Just (Matcher (argBefore args) (argAfter args) (argMax args) (not . f))
+          | otherwise      = Just (Matcher (argBefore args) (argAfter args) (argMax args) f)
 
 -- | Compute the command and arguments currently in the textbox.
 clientActiveCommand ::
@@ -781,8 +819,13 @@ addConnection attempts lastTime stsUpgrade network st =
             -- don't bother delaying on the first reconnect
             let delay = 15 * max 0 (attempts - 1)
             c <- createConnection delay settings1
-            seed <- Random.withSystemRandom (Random.asGenIO Random.save)
-            let cs = newNetworkState network settings1 c (PingConnecting attempts lastTime) seed
+            seed <- Random.newStdGen
+            let restrict = case view ssTls settings1 of
+                             TlsStart -> StartTLSRestriction
+                             TlsYes   -> WaitTLSRestriction
+                             TlsNo    -> NoRestriction
+                cs = newNetworkState network settings1 c
+                       (PingConnecting attempts lastTime restrict) seed
             traverse_ (sendMsg cs) (initialMessages cs)
             pure (set (clientConnections . at network) (Just cs) st)
 
@@ -791,7 +834,7 @@ applyStsPolicy stsUpgrade settings st =
   do now <- getCurrentTime
      let stsUpgrade'
            | Just{} <- stsUpgrade = stsUpgrade
-           | UseInsecure <- view ssTls settings
+           | TlsNo <- view ssTls settings
            , let host = Text.pack (view ssHostName settings)
            , Just policy <- view (clientStsPolicy . at host) st
            , now < view stsExpiration policy
@@ -799,7 +842,7 @@ applyStsPolicy stsUpgrade settings st =
            | otherwise = Nothing
      pure $ case stsUpgrade' of
               Just port -> set ssPort (Just (fromIntegral port))
-                         $ set ssTls UseTls settings
+                         $ set ssTls TlsYes settings
               Nothing   -> settings
 
 applyMessageToClientState ::
@@ -808,29 +851,13 @@ applyMessageToClientState ::
   Text                       {- ^ network name             -} ->
   NetworkState               {- ^ network connection state -} ->
   ClientState                {- ^ client state             -} ->
-  ([RawIrcMsg], Maybe DCCUpdate, ClientState) {- ^ response , DCC updates, updated state -}
+  ([RawIrcMsg], ClientState) {- ^ response , DCC updates, updated state -}
 applyMessageToClientState time irc network cs st =
-  cs' `seq` (reply, dccUp, st')
+  cs' `seq` (reply, st')
   where
-    (reply, cs') = applyMessage time irc cs
-    (st', dccUp) = queueDCCTransfer network irc
-                 $ applyWindowRenames network irc
-                 $ set (clientConnections . ix network) cs' st
-
--- | Queue a DCC transfer when the message is correct. Await for user
---   confirmation to start the download.
-queueDCCTransfer :: Text -> IrcMsg -> ClientState
-                 -> (ClientState, Maybe DCCUpdate)
-queueDCCTransfer network ctcpMsg st
-  | Just (fromU, _target, command, txt) <- ctcpToTuple ctcpMsg
-  , command == "DCC", dccState <- view clientDCC st
-  = case (parseSEND network fromU txt, parseACCEPT dccState fromU txt) of
-      (Right offer, _) -> (set clientDCC (insertAsNewMax offer dccState) st, Nothing)
-      (_, Just upd) -> (st, Just upd)
-      (_, _) -> (st, Nothing)
-
-  | otherwise = (st, Nothing)
-
+    Apply reply cs' = applyMessage time irc cs
+    st' = applyWindowRenames network irc
+        $ set (clientConnections . ix network) cs' st
 
 -- | When a nick change happens and there is an open query window for that nick
 -- and there isn't an open query window for the new nick, rename the window.
@@ -844,7 +871,7 @@ applyWindowRenames network (Nick old new) st
                         $ over clientWindows moveWindow st
   | otherwise = st
   where
-    old' = userNick old
+    old' = userNick (srcUser old)
 
     mkFocus = ChannelFocus network
 
@@ -894,20 +921,22 @@ jumpToActivity st =
         Just focus -> changeFocus focus st
         Nothing    -> st
   where
-    windowList   = views clientWindows Map.toAscList st
+    windowList   = filter (not . view winSilent . snd)
+                 $ views clientWindows Map.toAscList st
     highPriority = find (\x -> WLImportant == view winMention (snd x)) windowList
     lowPriority  = find (\x -> view winUnread (snd x) > 0) windowList
 
--- | Jump the focus directly to a window based on its zero-based index.
+-- | Jump the focus directly to a window based on its zero-based index
+-- while ignoring hidden windows.
 jumpFocus ::
-  Int {- ^ zero-based window index -} ->
+  Char {- ^ window name -} ->
   ClientState -> ClientState
-jumpFocus i st
-  | 0 <= i, i < Map.size windows = changeFocus focus st
-  | otherwise                    = st
+jumpFocus i st =
+  case find p (Map.assocs (view clientWindows st)) of
+    Nothing        -> st
+    Just (focus,_) -> changeFocus focus st
   where
-    windows   = view clientWindows st
-    (focus,_) = Map.elemAt i windows
+    p (_, w) = view winName w == Just i
 
 
 -- | Change the window focus to the given value, reset the subfocus
@@ -1018,29 +1047,40 @@ stepFocus selector st =
     Just k  -> changeFocus k st
     Nothing -> st
   where
-    (l,r) = Map.split (view clientFocus st) (view clientWindows st)
+    (l,r) = Map.split (view clientFocus st)
+          $ Map.filter (views winHidden not)
+          $ view clientWindows st
 
--- | Compute the set of extra identifiers that should be highlighted given
--- a particular network state.
-clientHighlights ::
-  NetworkState       {- ^ network state               -} ->
-  ClientState        {- ^ client state                -} ->
-  HashSet Identifier {- ^ extra highlight identifiers -}
-clientHighlights cs st =
-  HashSet.insert
-    (view csNick cs)
-    (view (clientConfig . configExtraHighlights) st)
+clientHighlightsFocus ::
+  Focus ->
+  ClientState ->
+  HashMap Identifier Highlight
+clientHighlightsFocus focus st =
+  case focus of
+    ChannelFocus n c -> netcase n (Just c)
+    NetworkFocus n   -> netcase n Nothing
+    Unfocused        -> base
+  where
+    base = HashMap.fromList [(x, HighlightMe) | x <- view (clientConfig . configExtraHighlights) st]
+        <> HashMap.fromList [(x, HighlightNone) | x <- view (clientConfig . configNeverHighlights) st]
+        <> view clientHighlights st
 
--- | Compute the set of extra identifiers that should be highlighted given
--- a particular network.
-clientHighlightsNetwork ::
-  Text               {- ^ network                     -} ->
-  ClientState        {- ^ client state                -} ->
-  HashSet Identifier {- ^ extra highlight identifiers -}
-clientHighlightsNetwork network st =
-  case preview (clientConnection network) st of
-    Just cs -> clientHighlights cs st
-    Nothing -> view (clientConfig . configExtraHighlights) st
+    replace x y =
+      case x of
+        HighlightError -> y
+        _              -> x
+
+    netcase n mbC =
+      case preview (clientConnection n) st of
+        Nothing -> view clientHighlights st
+        Just cs ->
+          HashMap.unionWith
+            replace
+            (HashMap.insert (view csNick cs) HighlightMe base)
+            (HashMap.fromList [(u, HighlightNick)
+                                | Just c <- [mbC]
+                                , u <- views (csChannels . ix c . chanUsers) HashMap.keys cs
+                                , Text.length (idText u) > 1 ])
 
 -- | Produce the list of window names configured for the client.
 clientWindowNames ::
