@@ -21,6 +21,7 @@ module Client.State.Network
   -- * Connection state
     NetworkState(..)
   , AuthenticateState(..)
+  , ConnectRestriction(..)
   , newNetworkState
 
   -- * Lenses
@@ -47,6 +48,7 @@ module Client.State.Network
   , csMessageHooks
   , csAuthenticationState
   , csSeed
+  , csAway
 
   -- * Cross-message state
   , Transaction(..)
@@ -58,9 +60,13 @@ module Client.State.Network
   -- * Messages interactions
   , sendMsg
   , initialMessages
-  , applyMessage
   , squelchIrcMsg
 
+  -- * NetworkState update
+  , Apply(..)
+  , applyMessage
+  , hideMessage
+
   -- * Timer information
   , PingStatus(..)
   , _PingConnecting
@@ -75,6 +81,8 @@ module Client.State.Network
   ) where
 
 import qualified Client.Authentication.Ecdsa as Ecdsa
+import qualified Client.Authentication.Ecdh as Ecdh
+import qualified Client.Authentication.Scram as Scram
 import           Client.Configuration.ServerSettings
 import           Client.Network.Async
 import           Client.State.Channel
@@ -82,7 +90,6 @@ import           Client.UserHost
 import           Client.Hook (MessageHook)
 import           Client.Hooks (messageHooks)
 import           Control.Lens
-import qualified Control.Monad.ST as ST
 import           Data.HashMap.Strict (HashMap)
 import qualified Data.HashMap.Strict as HashMap
 import qualified Data.HashSet as HashSet
@@ -107,7 +114,8 @@ import           Irc.Modes
 import           Irc.RawIrcMsg
 import           Irc.UserInfo
 import           LensUtils
-import qualified System.Random.MWC as Random
+import qualified System.Random as Random
+import qualified Data.ByteString.Base64 as B64
 
 -- | State tracked for each IRC connection
 data NetworkState = NetworkState
@@ -127,6 +135,7 @@ data NetworkState = NetworkState
   , _csNetwork      :: !Text -- ^ name of network connection
   , _csMessageHooks :: ![MessageHook] -- ^ names of message hooks to apply to this connection
   , _csAuthenticationState :: !AuthenticateState
+  , _csAway         :: !Bool -- ^ Tracks when you are marked away
 
   -- Timing information
   , _csNextPingTime :: !(Maybe UTCTime) -- ^ time for next ping event
@@ -136,7 +145,7 @@ data NetworkState = NetworkState
   , _csCertificate  :: ![Text]
 
   -- Randomization
-  , _csSeed         :: Random.Seed
+  , _csSeed         :: Random.StdGen
   }
 
 -- | State of the authentication transaction
@@ -146,13 +155,23 @@ data AuthenticateState
   | AS_EcdsaStarted       -- ^ ECDSA-NIST mode initiated
   | AS_EcdsaWaitChallenge -- ^ ECDSA-NIST user sent waiting for challenge
   | AS_ExternalStarted    -- ^ EXTERNAL mode initiated
-  deriving Show
+  | AS_ScramStarted
+  | AS_Scram1 Scram.Phase1
+  | AS_Scram2 Scram.Phase2
+  | AS_EcdhStarted
+  | AS_EcdhWaitChallenge Ecdh.Phase1
 
 -- | Status of the ping timer
 data PingStatus
   = PingSent !UTCTime -- ^ ping sent at given time, waiting for pong
   | PingNone          -- ^ not waiting for a pong
-  | PingConnecting !Int !(Maybe UTCTime) -- ^ number of attempts, last known connection time
+  | PingConnecting !Int !(Maybe UTCTime) !ConnectRestriction -- ^ number of attempts, last known connection time
+  deriving Show
+
+data ConnectRestriction
+  = NoRestriction       -- ^ no message restriction
+  | StartTLSRestriction -- ^ STARTTLS hasn't finished
+  | WaitTLSRestriction  -- ^ No messages allowed until TLS starts
   deriving Show
 
 -- | Timer-based events
@@ -247,7 +266,7 @@ newNetworkState ::
   ServerSettings    {- ^ server settings           -} ->
   NetworkConnection {- ^ active network connection -} ->
   PingStatus        {- ^ initial ping status       -} ->
-  Random.Seed       {- ^ initial random seed       -} ->
+  Random.StdGen     {- ^ initial random seed       -} ->
   NetworkState      {- ^ new network state         -}
 newNetworkState network settings sock ping seed = NetworkState
   { _csUserInfo     = UserInfo "*" "" ""
@@ -266,6 +285,7 @@ newNetworkState network settings sock ping seed = NetworkState
   , _csNetwork      = network
   , _csMessageHooks = buildMessageHooks (view ssMessageHooks settings)
   , _csAuthenticationState = AS_None
+  , _csAway         = False
   , _csPingStatus   = ping
   , _csLatency      = Nothing
   , _csNextPingTime = Nothing
@@ -279,11 +299,25 @@ buildMessageHooks = mapMaybe \(HookConfig name args) ->
   do hookFun <- HashMap.lookup name messageHooks
      hookFun args
 
+data Apply = Apply [RawIrcMsg] NetworkState
+
+hideMessage :: IrcMsg -> Bool
+hideMessage m =
+  case m of
+    Authenticate{} -> True
+    BatchStart{} -> True
+    BatchEnd{} -> True
+    Ping{} -> True
+    Pong{} -> True
+    Reply _ RPL_WHOSPCRPL [_,"616",_,_,_,_] -> True
+    _ -> False
+
 -- | Used for updates to a 'NetworkState' that require no reply.
---
--- @noReply x = ([], x)@
-noReply :: NetworkState -> ([RawIrcMsg], NetworkState)
-noReply x = ([], x)
+noReply :: NetworkState -> Apply
+noReply = reply []
+
+reply :: [RawIrcMsg] -> NetworkState -> Apply
+reply = Apply
 
 overChannel :: Identifier -> (ChannelState -> ChannelState) -> NetworkState -> NetworkState
 overChannel chan = overStrict (csChannels . ix chan)
@@ -291,76 +325,77 @@ overChannel chan = overStrict (csChannels . ix chan)
 overChannels :: (ChannelState -> ChannelState) -> NetworkState -> NetworkState
 overChannels = overStrict (csChannels . traverse)
 
-applyMessage :: ZonedTime -> IrcMsg -> NetworkState -> ([RawIrcMsg], NetworkState)
+applyMessage :: ZonedTime -> IrcMsg -> NetworkState -> Apply
 applyMessage msgWhen msg cs
   = applyMessage' msgWhen msg
   $ set csLastReceived (Just $! zonedTimeToUTC msgWhen) cs
 
-applyMessage' :: ZonedTime -> IrcMsg -> NetworkState -> ([RawIrcMsg], NetworkState)
+applyMessage' :: ZonedTime -> IrcMsg -> NetworkState -> Apply
 applyMessage' msgWhen msg cs =
   case msg of
-    Ping args -> ([ircPong args], cs)
-    Pong _    -> noReply $ doPong msgWhen cs
+    Ping args -> reply [ircPong args] cs
+    Pong _    -> noReply (doPong msgWhen cs)
     Join user chan acct _ ->
-         ( reply
-         , recordUser user acct
-         $ overChannel chan (joinChannel (userNick user))
-         $ createOnJoin user chan cs )
+         reply response
+         $ recordUser (srcUser user) acct
+         $ overChannel chan (joinChannel (userNick (srcUser user)))
+         $ createOnJoin (srcUser user) chan cs
      where
-       showAccounts = view (csSettings . ssShowAccounts) cs
-       reply
-         | userNick user == view csNick cs =
-              ircMode chan [] :
-              [ircWho [idText chan, "%tuhna,616"] | showAccounts ]
-         | otherwise = []
+       response =
+         [ircMode chan [] | userNick (srcUser user) == view csNick cs]
 
     Account user acct ->
            noReply
-         $ recordUser user acct cs
+         $ recordUser (srcUser user) acct cs
 
     Chghost user newUser newHost ->
            noReply
-         $ updateUserInfo (userNick user) newUser newHost cs
+         $ updateUserInfo (userNick (srcUser user)) newUser newHost cs
 
     Quit user _reason ->
            noReply
-         $ forgetUser (userNick user)
-         $ overChannels (partChannel (userNick user)) cs
+         $ forgetUser (userNick (srcUser user))
+         $ overChannels (partChannel (userNick (srcUser user))) cs
 
-    Part user chan _mbreason -> exitChannel chan (userNick user)
+    Part user chan _mbreason -> exitChannel chan (userNick (srcUser user))
 
     Kick _kicker chan nick _reason -> exitChannel chan nick
 
     Nick oldNick newNick ->
+         let nick = userNick (srcUser oldNick) in
            noReply
-         $ renameUser (userNick oldNick) newNick
-         $ updateMyNick (userNick oldNick) newNick
-         $ overChannels (nickChange (userNick oldNick) newNick) cs
+         $ renameUser nick newNick
+         $ updateMyNick nick newNick
+         $ overChannels (nickChange nick newNick) cs
 
-    Reply RPL_WELCOME (me:_) -> doWelcome msgWhen (mkId me) cs
-    Reply RPL_SASLSUCCESS _ -> ([ircCapEnd], cs)
-    Reply RPL_SASLFAIL _ -> ([ircCapEnd], cs)
+    Reply _ RPL_WELCOME (me:_) -> doWelcome msgWhen (mkId me) cs
+    Reply _ RPL_SASLSUCCESS _ -> reply [ircCapEnd] cs
+    Reply _ ERR_SASLFAIL _ -> reply [ircCapEnd] cs
+    Reply _ ERR_SASLABORTED _ -> reply [ircCapEnd] cs
+    Reply _ RPL_SASLMECHS _ -> reply [ircCapEnd] cs
 
-    Reply ERR_NICKNAMEINUSE (_:badnick:_)
+    Reply _ ERR_NICKNAMEINUSE (_:badnick:_)
       | PingConnecting{} <- view csPingStatus cs -> doBadNick badnick cs
-    Reply ERR_BANNEDNICK (_:badnick:_)
+    Reply _ ERR_BANNEDNICK (_:badnick:_)
       | PingConnecting{} <- view csPingStatus cs -> doBadNick badnick cs
-    Reply ERR_ERRONEUSNICKNAME (_:badnick:_)
+    Reply _ ERR_ERRONEUSNICKNAME (_:badnick:_)
+      | PingConnecting{} <- view csPingStatus cs -> doBadNick badnick cs
+    Reply _ ERR_UNAVAILRESOURCE (_:badnick:_)
       | PingConnecting{} <- view csPingStatus cs -> doBadNick badnick cs
 
-    Reply RPL_HOSTHIDDEN (_:host:_) ->
+    Reply _ RPL_HOSTHIDDEN (_:host:_) ->
         noReply (set (csUserInfo . uiHost) host cs)
 
     -- /who <#channel> %tuhna,616
-    Reply RPL_WHOSPCRPL [_me,"616",user,host,nick,acct] ->
+    Reply _ RPL_WHOSPCRPL [_me,"616",user,host,nick,acct] ->
        let acct' = if acct == "0" then "*" else acct
        in noReply (recordUser (UserInfo (mkId nick) user host) acct' cs)
 
-    Reply code args        -> noReply (doRpl code msgWhen args cs)
+    Reply _ code args      -> doRpl code msgWhen args cs
     Cap cmd                -> doCap cmd cs
     Authenticate param     -> doAuthenticate param cs
-    Mode who target (modes:params)  -> doMode msgWhen who target modes params cs
-    Topic user chan topic  -> noReply (doTopic msgWhen user chan topic cs)
+    Mode who target (modes:params) -> doMode msgWhen (srcUser who) target modes params cs
+    Topic user chan topic  -> noReply (doTopic msgWhen (srcUser user) chan topic cs)
     _                      -> noReply cs
   where
     exitChannel chan nick
@@ -384,7 +419,8 @@ pruneUsers cs = over csUsers (`HashMap.intersection` u) cs
 doWelcome ::
   ZonedTime  {- ^ message received -} ->
   Identifier {- ^ my nickname      -} ->
-  NetworkState -> ([RawIrcMsg], NetworkState)
+  NetworkState ->
+  Apply
 doWelcome msgWhen me
   = noReply
   . set csNick me
@@ -395,26 +431,23 @@ doWelcome msgWhen me
 doBadNick ::
   Text {- ^ bad nickname -} ->
   NetworkState ->
-  ([RawIrcMsg], NetworkState) {- ^ replies, updated state -}
+  Apply
 doBadNick badNick cs =
   case NonEmpty.dropWhile (badNick/=) (view (csSettings . ssNicks) cs) of
-    _:next:_ -> ([ircNick next], cs)
+    _:next:_ -> reply [ircNick next] cs
     _        -> doRandomNick cs
 
 -- | Pick a random nickname now that we've run out of choices
-doRandomNick :: NetworkState -> ([RawIrcMsg], NetworkState)
-doRandomNick cs = ([ircNick candidate], cs')
+doRandomNick :: NetworkState -> Apply
+doRandomNick cs = reply [ircNick candidate] cs'
   where
     limit       = 9 -- RFC 2812 puts the maximum nickname length as low as 9!
     range       = (0, 99999::Int) -- up to 5 random digits
     suffix      = show n
     primaryNick = NonEmpty.head (view (csSettings . ssNicks) cs)
     candidate   = Text.take (limit-length suffix) primaryNick <> Text.pack suffix
-    cs'         = set csSeed seed' cs
 
-    (n, seed')  = ST.runST (do gen <- Random.restore (view csSeed cs)
-                               (,) <$> Random.uniformR range gen <*> Random.save gen
-                           )
+    (n, cs')    = cs & csSeed %%~ Random.randomR range
 
 doTopic :: ZonedTime -> UserInfo -> Identifier -> Text -> NetworkState -> NetworkState
 doTopic when user chan topic =
@@ -432,36 +465,38 @@ parseTimeParam txt =
       Just $! posixSecondsToUTCTime (fromInteger i)
     _ -> Nothing
 
-doRpl :: ReplyCode -> ZonedTime -> [Text] -> NetworkState -> NetworkState
-doRpl cmd msgWhen args =
+doRpl :: ReplyCode -> ZonedTime -> [Text] -> NetworkState -> Apply
+doRpl cmd msgWhen args cs =
   case cmd of
     RPL_UMODEIS ->
-      \ns ->
       case args of
         _me:modes:params
-          | Just xs <- splitModes (view csUmodeTypes ns) modes params ->
-                 doMyModes xs
-               $ set csModes "" ns -- reset modes
-        _ -> ns
+          | Just xs <- splitModes (view csUmodeTypes cs) modes params ->
+                 noReply
+               $ doMyModes xs
+               $ set csModes "" cs -- reset modes
+        _ -> noReply cs
 
     RPL_SNOMASK ->
       case args of
         _me:snomask0:_
           | Just snomask <- Text.stripPrefix "+" snomask0 ->
-           set csSnomask (Text.unpack snomask)
-        _             -> id
+           noReply (set csSnomask (Text.unpack snomask) cs)
+        _ -> noReply cs
 
     RPL_NOTOPIC ->
       case args of
-        _me:chan:_ -> overChannel
+        _me:chan:_ -> noReply
+                    $ overChannel
                         (mkId chan)
                         (setTopic "" . set chanTopicProvenance Nothing)
-        _          -> id
+                        cs
+        _ -> noReply cs
 
     RPL_TOPIC ->
       case args of
-        _me:chan:topic:_ -> overChannel (mkId chan) (setTopic topic)
-        _                -> id
+        _me:chan:topic:_ -> noReply (overChannel (mkId chan) (setTopic topic) cs)
+        _                -> noReply cs
 
     RPL_TOPICWHOTIME ->
       case args of
@@ -470,99 +505,107 @@ doRpl cmd msgWhen args =
                        { _topicAuthor = parseUserInfo who
                        , _topicTime   = when
                        }
-          in overChannel (mkId chan) (set chanTopicProvenance (Just prov))
-        _ -> id
+          in noReply (overChannel (mkId chan) (set chanTopicProvenance (Just prov)) cs)
+        _ -> noReply cs
 
     RPL_CREATIONTIME ->
       case args of
         _me:chan:whenTxt:_ | Just when <- parseTimeParam whenTxt ->
-          overChannel (mkId chan) (set chanCreation (Just when))
-        _ -> id
+          noReply (overChannel (mkId chan) (set chanCreation (Just when)) cs)
+        _ -> noReply cs
 
     RPL_CHANNEL_URL ->
       case args of
         _me:chan:urlTxt:_ ->
-          overChannel (mkId chan) (set chanUrl (Just urlTxt))
-        _ -> id
+          noReply (overChannel (mkId chan) (set chanUrl (Just urlTxt)) cs)
+        _ -> noReply cs
 
-    RPL_MYINFO -> myinfo args
+    RPL_MYINFO -> noReply (myinfo args cs)
 
-    RPL_ISUPPORT -> isupport args
+    RPL_ISUPPORT -> noReply (isupport args cs)
 
     RPL_NAMREPLY ->
       case args of
         _me:_sym:_tgt:x:_ ->
+           noReply $
            over csTransaction
                 (\t -> let xs = view _NamesTransaction t
                        in xs `seq` NamesTransaction (x:xs))
-        _ -> id
+                cs
+        _ -> noReply cs
 
     RPL_ENDOFNAMES ->
       case args of
-        _me:tgt:_ -> loadNamesList (mkId tgt)
-        _         -> id
+        _me:tgt:_ -> noReply (loadNamesList (mkId tgt) cs)
+        _         -> noReply cs
 
     RPL_BANLIST ->
       case args of
-        _me:_tgt:mask:who:whenTxt:_ -> recordListEntry mask who whenTxt
-        _                           -> id
+        _me:_tgt:mask:who:whenTxt:_ -> noReply (recordListEntry mask who whenTxt cs)
+        _                           -> noReply cs
 
     RPL_ENDOFBANLIST ->
       case args of
-        _me:tgt:_ -> saveList 'b' tgt
-        _         -> id
+        _me:tgt:_ -> noReply (saveList 'b' tgt cs)
+        _         -> noReply cs
 
     RPL_QUIETLIST ->
       case args of
-        _me:_tgt:_q:mask:who:whenTxt:_ -> recordListEntry mask who whenTxt
-        _                              -> id
+        _me:_tgt:_q:mask:who:whenTxt:_ -> noReply (recordListEntry mask who whenTxt cs)
+        _                              -> noReply cs
 
     RPL_ENDOFQUIETLIST ->
       case args of
-        _me:tgt:_ -> saveList 'q' tgt
-        _         -> id
+        _me:tgt:_ -> noReply (saveList 'q' tgt cs)
+        _         -> noReply cs
 
     RPL_INVEXLIST ->
       case args of
-        _me:_tgt:mask:who:whenTxt:_ -> recordListEntry mask who whenTxt
-        _                           -> id
+        _me:_tgt:mask:who:whenTxt:_ -> noReply (recordListEntry mask who whenTxt cs)
+        _                           -> noReply cs
 
     RPL_ENDOFINVEXLIST ->
       case args of
-        _me:tgt:_ -> saveList 'I' tgt
-        _         -> id
+        _me:tgt:_ -> noReply (saveList 'I' tgt cs)
+        _         -> noReply cs
 
     RPL_EXCEPTLIST ->
       case args of
-        _me:_tgt:mask:who:whenTxt:_ -> recordListEntry mask who whenTxt
-        _                           -> id
+        _me:_tgt:mask:who:whenTxt:_ -> noReply (recordListEntry mask who whenTxt cs)
+        _                           -> noReply cs
 
     RPL_ENDOFEXCEPTLIST ->
       case args of
-        _me:tgt:_ -> saveList 'e' tgt
-        _         -> id
+        _me:tgt:_ -> noReply (saveList 'e' tgt cs)
+        _         -> noReply cs
 
     RPL_WHOREPLY ->
       case args of
         _me:_tgt:uname:host:_server:nick:_ ->
-          over csTransaction $ \t ->
+          noReply $
+          over csTransaction (\t ->
             let !x  = UserInfo (mkId nick) uname host
                 !xs = view _WhoTransaction t
-            in WhoTransaction (x : xs)
-        _ -> id
+            in WhoTransaction (x : xs))
+            cs
+        _ -> noReply cs
 
-    RPL_ENDOFWHO -> massRegistration
+    RPL_ENDOFWHO -> noReply (massRegistration cs)
 
     RPL_CHANNELMODEIS ->
       case args of
         _me:chan:modes:params ->
-              snd -- channel mode reply shouldn't trigger messages
-            . doMode msgWhen who chanId modes params
-            . set (csChannels . ix chanId . chanModes) Map.empty
+              doMode msgWhen who chanId modes params
+            $ set (csChannels . ix chanId . chanModes) Map.empty cs
             where chanId = mkId chan
                   !who = UserInfo "*" "" ""
-        _ -> id
-    _ -> id
+        _ -> noReply cs
+
+    -- Away flag tracking
+    RPL_NOWAWAY -> noReply (set csAway True cs)
+    RPL_UNAWAY  -> noReply (set csAway False cs)
+
+    _ -> noReply cs
 
 
 -- | Add an entry to a mode list transaction
@@ -632,8 +675,8 @@ squelchReply rpl =
 -- so the user shouldn't need to see them directly to get the
 -- relevant information.
 squelchIrcMsg :: IrcMsg -> Bool
-squelchIrcMsg (Reply rpl _) = squelchReply rpl
-squelchIrcMsg _             = False
+squelchIrcMsg (Reply _ rpl _) = squelchReply rpl
+squelchIrcMsg _               = False
 
 doMode ::
   ZonedTime {- ^ time of message -} ->
@@ -641,23 +684,21 @@ doMode ::
   Identifier {- ^ channel        -} ->
   Text       {- ^ mode flags     -} ->
   [Text]     {- ^ mode parameters -} ->
-  NetworkState -> ([RawIrcMsg], NetworkState)
+  NetworkState ->
+  Apply
 doMode when who target modes args cs
   | view csNick cs == target
   , Just xs <- splitModes (view csUmodeTypes cs) modes args =
         noReply (doMyModes xs cs)
 
   | isChannelIdentifier cs target
-  , Just xs <- splitModes (view csModeTypes cs) modes args =
-        let cs' = doChannelModes when who target xs cs
+  , Just xs <- splitModes (view csModeTypes cs) modes args
+  , let cs' = doChannelModes when who target xs cs =
 
-            finish | iHaveOp target cs' = popQueue
-                   | otherwise          = noReply
-
-        in finish cs'
-  where
-    popQueue :: NetworkState -> ([RawIrcMsg], NetworkState)
-    popQueue = csChannels . ix target . chanQueuedModeration <<.~ []
+    if iHaveOp target cs'
+      then let (response, cs_) = cs' & csChannels . ix target . chanQueuedModeration <<.~ []
+           in reply response cs_
+      else noReply cs'
 
 doMode _ _ _ _ _ cs = noReply cs -- ignore bad mode command
 
@@ -726,7 +767,8 @@ selectCaps cs offered = (supported `intersect` Map.keys capMap)
       sasl ++ serverTime ++
       ["multi-prefix", "batch", "znc.in/playback", "znc.in/self-message"
       , "cap-notify", "extended-join", "account-notify", "chghost"
-      , "userhost-in-names", "account-tag" ]
+      , "userhost-in-names", "account-tag", "solanum.chat/identify-msg"
+      , "solanum.chat/realhost" ]
 
     -- logic for using IRCv3.2 server-time if available and falling back
     -- to ZNC's specific extension otherwise.
@@ -738,39 +780,110 @@ selectCaps cs offered = (supported `intersect` Map.keys capMap)
     ss = view csSettings cs
     sasl = ["sasl" | isJust (view ssSaslMechanism ss) ]
 
-doAuthenticate :: Text -> NetworkState -> ([RawIrcMsg], NetworkState)
-doAuthenticate param cs =
+decodeAuthParam :: Text -> Maybe B.ByteString
+decodeAuthParam "+" = Just ""
+decodeAuthParam xs =
+  case B64.decode (Text.encodeUtf8 xs) of
+    Right bs -> Just bs
+    Left _ -> Nothing
+
+abortAuth :: NetworkState -> Apply
+abortAuth = reply [ircAuthenticate "*"] . set csAuthenticationState AS_None
+
+doAuthenticate :: Text -> NetworkState -> Apply
+doAuthenticate paramTxt cs =
+  case decodeAuthParam paramTxt of
+    Nothing -> abortAuth cs
+    Just param -> doAuthenticate' param cs
+
+doAuthenticate' :: B.ByteString -> NetworkState -> Apply
+doAuthenticate' param cs =
   case view csAuthenticationState cs of
     AS_PlainStarted
-      | "+" <- param
+      | B.null param
       , Just (SaslPlain mbAuthz authc (SecretText pass)) <- view ssSaslMechanism ss
       , let authz = fromMaybe "" mbAuthz
-      -> (ircAuthenticates (encodePlainAuthentication authz authc pass),
-          set csAuthenticationState AS_None cs)
+      -> reply
+           (ircAuthenticates (encodePlainAuthentication authz authc pass))
+           (set csAuthenticationState AS_None cs)
 
     AS_ExternalStarted
-      | "+" <- param
+      | B.null param
       , Just (SaslExternal mbAuthz) <- view ssSaslMechanism ss
       , let authz = fromMaybe "" mbAuthz
-      -> (ircAuthenticates (encodeExternalAuthentication authz),
-          set csAuthenticationState AS_None cs)
+      -> reply
+           (ircAuthenticates (encodeExternalAuthentication authz))
+           (set csAuthenticationState AS_None cs)
 
     AS_EcdsaStarted
-      | "+" <- param
+      | B.null param
       , Just (SaslEcdsa mbAuthz authc _) <- view ssSaslMechanism ss
-      , let authz = fromMaybe authc mbAuthz
-      -> (ircAuthenticates (Ecdsa.encodeAuthentication authz authc),
-          set csAuthenticationState AS_EcdsaWaitChallenge cs)
+      -> reply
+           (ircAuthenticates (Ecdsa.encodeAuthentication mbAuthz authc))
+           (set csAuthenticationState AS_EcdsaWaitChallenge cs)
 
-    AS_EcdsaWaitChallenge -> ([], cs) -- handled in Client.EventLoop!
+    AS_EcdsaWaitChallenge -> noReply cs -- handled in Client.EventLoop!
 
-    _ -> ([ircCapEnd], cs) -- really shouldn't happen
+    AS_ScramStarted
+      | B.null param
+      , Just (SaslScram digest mbAuthz user (SecretText pass))
+          <- view ssSaslMechanism ss
+      , let authz = fromMaybe "" mbAuthz
+      , (nonce, cs') <- cs & csSeed %%~ scramNonce
+      , (msg, scram1) <-
+          Scram.initiateScram digest
+            (Text.encodeUtf8 user)
+            (Text.encodeUtf8 authz)
+            (Text.encodeUtf8 pass)
+            nonce
+      -> reply
+           (ircAuthenticates msg)
+           (set csAuthenticationState (AS_Scram1 scram1) cs')
+
+    AS_Scram1 scram1
+      | Just (rsp, scram2) <- Scram.addServerFirst scram1 param
+      -> reply
+           (ircAuthenticates rsp)
+           (set csAuthenticationState (AS_Scram2 scram2) cs)
+
+    AS_Scram2 scram2
+      | Scram.addServerFinal scram2 param
+      -> reply
+           [ircAuthenticate "+"]
+           (set csAuthenticationState AS_None cs)
+
+    AS_EcdhStarted
+      | B.null param
+      , Just (SaslEcdh mbAuthz authc (SecretText key)) <- view ssSaslMechanism ss
+      , Just (rsp, ecdh1) <- Ecdh.clientFirst mbAuthz authc key
+      -> reply
+           (ircAuthenticates rsp)
+           (set csAuthenticationState (AS_EcdhWaitChallenge ecdh1) cs)
+    
+    AS_EcdhWaitChallenge ecdh1
+      | Just rsp <- Ecdh.clientResponse ecdh1 param
+      -> reply (ircAuthenticates rsp) (set csAuthenticationState AS_None cs)
+
+    _ -> abortAuth cs
 
   where
     ss = view csSettings cs
 
+scramNonce :: Random.StdGen -> (B.ByteString, Random.StdGen)
+scramNonce = go [] nonceSize
+  where
+    alphabet = "!\"#$%&'()*+-./0123456789:;<=>?@\
+               \ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`\
+               \abcdefghijklmnopqrstuvwxyz{|}~"
 
-doCap :: CapCmd -> NetworkState -> ([RawIrcMsg], NetworkState)
+    nonceSize = 20 :: Int -- ceiling (128 / logBase 9 (length alphabet))
+
+    go acc 0 g = (B.pack acc, g)
+    go acc i g =
+      case Random.randomR (0, B.length alphabet-1) g of
+        (x,g') -> go (B.index alphabet x:acc) (i-1) g'
+
+doCap :: CapCmd -> NetworkState -> Apply
 doCap cmd cs =
   case cmd of
     (CapLs CapMore caps) ->
@@ -779,19 +892,19 @@ doCap cmd cs =
         prevCaps = view (csTransaction . _CapLsTransaction) cs
 
     CapLs CapDone caps
-      | null reqCaps -> ([ircCapEnd], cs')
-      | otherwise    -> ([ircCapReq reqCaps], cs')
+      | null reqCaps -> reply [ircCapEnd] cs'
+      | otherwise    -> reply [ircCapReq reqCaps] cs'
       where
         reqCaps = selectCaps cs (caps ++ view (csTransaction . _CapLsTransaction) cs)
         cs' = set csTransaction NoTransaction cs
 
     CapNew caps
-      | null reqCaps -> ([], cs)
-      | otherwise    -> ([ircCapReq reqCaps], cs)
+      | null reqCaps -> noReply cs
+      | otherwise    -> reply [ircCapReq reqCaps] cs
       where
         reqCaps = selectCaps cs caps
 
-    CapDel _ -> ([],cs)
+    CapDel _ -> noReply cs
 
     CapAck caps
       | let ss = view csSettings cs
@@ -799,16 +912,22 @@ doCap cmd cs =
       , Just mech <- view ssSaslMechanism ss ->
         case mech of
           SaslEcdsa{} ->
-            ([ircAuthenticate Ecdsa.authenticationMode],
-             set csAuthenticationState AS_EcdsaStarted cs)
+            reply [ircAuthenticate Ecdsa.authenticationMode]
+                  (set csAuthenticationState AS_EcdsaStarted cs)
           SaslPlain{} ->
-            ([ircAuthenticate "PLAIN"],
-             set csAuthenticationState AS_PlainStarted cs)
+            reply [ircAuthenticate "PLAIN"]
+                  (set csAuthenticationState AS_PlainStarted cs)
           SaslExternal{} ->
-            ([ircAuthenticate "EXTERNAL"],
-             set csAuthenticationState AS_ExternalStarted cs)
+            reply [ircAuthenticate "EXTERNAL"]
+                  (set csAuthenticationState AS_ExternalStarted cs)
+          SaslScram digest _ _ _ ->
+            reply [ircAuthenticate (Scram.mechanismName digest)]
+                  (set csAuthenticationState AS_ScramStarted cs)
+          SaslEcdh{} ->
+            reply [ircAuthenticate Ecdh.mechanismName]
+                  (set csAuthenticationState AS_EcdhStarted cs)
 
-    _ -> ([ircCapEnd], cs)
+    _ -> reply [ircCapEnd] cs
 
 initialMessages :: NetworkState -> [RawIrcMsg]
 initialMessages cs
@@ -1020,7 +1139,9 @@ nextPingAction cs =
   where
     action =
       case view csPingStatus cs of
-        PingSent{}       -> TimedDisconnect
+        PingSent sentAt
+          | Just sentAt < view csLastReceived cs -> TimedSendPing
+          | otherwise -> TimedDisconnect
         PingNone         -> TimedSendPing
         PingConnecting{} -> TimedSendPing
 
