@@ -19,7 +19,6 @@ import           Client.State.Focus
 import           Client.View.Cert
 import           Client.View.ChannelInfo
 import           Client.View.Digraphs
-import           Client.View.DCCList
 import           Client.View.Help
 import           Client.View.IgnoreList
 import           Client.View.KeyMap
@@ -42,11 +41,10 @@ viewLines focus subfocus w !st =
       channelInfoImages network channel st
     (ChannelFocus network channel, FocusUsers)
       | view clientDetailView st -> userInfoImages network channel st
-      | otherwise                -> userListImages network channel st
+      | otherwise                -> userListImages network channel w st
     (ChannelFocus network channel, FocusMasks mode) ->
       maskListImages mode network channel w st
     (_, FocusWindows filt) -> windowsImages filt st
-    (_, FocusDCC)          -> dccImages st
     (_, FocusMentions)     -> mentionsViewLines w st
     (_, FocusPalette)      -> paletteViewLines pal
     (_, FocusDigraphs)     -> digraphLines w st
