@@ -1,3 +1,4 @@
+{-# Language OverloadedStrings #-}
 {-|
 Module      : Client.View.Windows
 Description : View of the list of open windows
@@ -20,23 +21,27 @@ import           Client.State.Window
 import           Client.State.Network
 import           Control.Lens
 import           Data.List
+import           Data.Maybe (fromMaybe)
 import qualified Data.Map as Map
 import           Graphics.Vty.Attributes
 import           Irc.Identifier
 
 -- | Draw the image lines associated with the @/windows@ command.
 windowsImages :: WindowsFilter -> ClientState -> [Image']
-windowsImages filt st = reverse (createColumns windows)
+windowsImages filt st
+  = reverse
+  $ createColumns
+  $ [ renderWindowColumns pal (char (view palError pal) 'h')    k v | (k,v) <- hiddenWindows ] ++
+    [ renderWindowColumns pal (char (view palWindowName pal) (name v)) k v | (k,v) <- windows ]
   where
-    windows = [ renderWindowColumns pal n k v
-              | (n,(k,v)) <- zip names
-                           $ views clientWindows Map.toAscList st
-              , windowMatcher filt st k
-              ]
-
-    pal     = clientPalette st
-    names   = clientWindowNames st ++ repeat '?'
+    pal = clientPalette st
+    name = fromMaybe ' ' . view winName
 
+    (hiddenWindows, windows)
+      = partition (view (_2 . winHidden))
+      $ filter (windowMatcher filt st . fst)
+      $ Map.toAscList
+      $ view clientWindows st
 
 ------------------------------------------------------------------------
 
@@ -61,9 +66,9 @@ windowMatcher _ _ _ = False
 ------------------------------------------------------------------------
 
 
-renderWindowColumns :: Palette -> Char -> Focus -> Window -> [Image']
+renderWindowColumns :: Palette -> Image' -> Focus -> Window -> [Image']
 renderWindowColumns pal name focus win =
-  [ char (view palWindowName pal) name
+  [ name
   , renderedFocus pal focus
   , renderedWindowInfo pal win
   ]
@@ -91,9 +96,9 @@ renderedFocus pal focus =
 
 renderedWindowInfo :: Palette -> Window -> Image'
 renderedWindowInfo pal win =
-  string (view newMsgAttrLens pal) (views winUnread show win) <>
-  char defAttr '/' <>
-  string (view palActivity pal) (views winTotal show win)
+  string (view newMsgAttrLens pal) (views winUnread show win) <> "/" <>
+  string (view palActivity    pal) (views winTotal  show win) <>
+  (if view winSilent win then text' (view palMeta pal) " silent" else mempty)
   where
     newMsgAttrLens =
       case view winMention win of
